from django import test
from ievv_opensource.ievv_sms import sms_registry
from unittest import mock


class TestRegistry(test.TestCase):
    def test_add_ok(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'mock'
        registry.add(mock_backend_class)
        self.assertIn('mock', registry._backend_class_map)

    def test_add_failed(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'mock'
        registry.add(mock_backend_class)
        with self.assertRaisesMessage(ValueError,
                                      "'mock' is already in the SMS backend registry."):
            registry.add(mock_backend_class)

    def test_remove(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'mock'
        registry.add(mock_backend_class)
        self.assertIn('mock', registry._backend_class_map)
        registry.remove(mock_backend_class)
        self.assertNotIn('mock', registry._backend_class_map)

    def test_contains(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'mock'
        registry.add(mock_backend_class)
        self.assertTrue(mock_backend_class in registry)

    def test_iter_backends(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class1 = mock.MagicMock()
        mock_backend_class1.get_backend_id.return_value = 'mock1'
        registry.add(mock_backend_class1)
        mock_backend_class2 = mock.MagicMock()
        mock_backend_class2.get_backend_id.return_value = 'mock2'
        registry.add(mock_backend_class2)
        self.assertEqual(set(registry.iter_backend_classes()), {
            mock_backend_class1, mock_backend_class2
        })

    def test__iter__(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class1 = mock.MagicMock()
        mock_backend_class1.get_backend_id.return_value = 'mock1'
        registry.add(mock_backend_class1)
        mock_backend_class2 = mock.MagicMock()
        mock_backend_class2.get_backend_id.return_value = 'mock2'
        registry.add(mock_backend_class2)
        self.assertEqual(set(iter(registry)), {
            mock_backend_class1, mock_backend_class2
        })

    @test.override_settings(IEVV_SMS_DEFAULT_BACKEND_ID=None)
    def test_get_default_backend_id__default(self):
        registry = sms_registry.MockableRegistry()
        self.assertEqual(registry.get_default_backend_id(), 'debugprint')

    @test.override_settings(IEVV_SMS_DEFAULT_BACKEND_ID='something')
    def test_get_default_backend_id__custom(self):
        registry = sms_registry.MockableRegistry()
        self.assertEqual(registry.get_default_backend_id(), 'something')

    def test_get_backend_class_by_id__ok(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'mock'
        registry.add(mock_backend_class)
        self.assertEqual(registry.get_backend_class_by_id('mock'),
                         mock_backend_class)

    def test_get_backend_class_by_id__invalid_backend_id(self):
        registry = sms_registry.MockableRegistry()
        with self.assertRaises(KeyError):
            registry.get_backend_class_by_id('invalid')

    @test.override_settings(IEVV_SMS_DEFAULT_BACKEND_ID='mock')
    def test_make_backend_instance__default_backend(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'mock'
        mock_backend_instance = mock.MagicMock()
        mock_backend_class.return_value = mock_backend_instance
        registry.add(mock_backend_class)
        backend = registry.make_backend_instance(
            phone_number='12345678',
            message='Testmessage'
        )
        mock_backend_class.assert_called_once_with(
            phone_number='12345678',
            message='Testmessage'
        )
        self.assertEqual(backend, mock_backend_instance)

    @test.override_settings(IEVV_SMS_DEFAULT_BACKEND_ID='mock')
    def test_make_backend_instance__custom_backend(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'custom'
        mock_backend_instance = mock.MagicMock()
        mock_backend_class.return_value = mock_backend_instance
        registry.add(mock_backend_class)
        backend = registry.make_backend_instance(
            phone_number='12345678',
            message='Testmessage',
            backend_id='custom'
        )
        mock_backend_class.assert_called_once_with(
            phone_number='12345678',
            message='Testmessage'
        )
        self.assertEqual(backend, mock_backend_instance)

    @test.override_settings(IEVV_SMS_DEFAULT_BACKEND_ID='mock')
    def test_send__default_backend(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'mock'
        mock_backend_instance = mock.MagicMock()
        mock_backend_class.return_value = mock_backend_instance
        registry.add(mock_backend_class)
        registry.send(
            phone_number='12345678',
            message='Testmessage'
        )
        mock_backend_class.assert_called_once_with(
            phone_number='12345678',
            message='Testmessage'
        )
        mock_backend_instance.send.assert_called_once_with()

    @test.override_settings(IEVV_SMS_DEFAULT_BACKEND_ID='mock')
    def test_send__custom_backend(self):
        registry = sms_registry.MockableRegistry()
        mock_backend_class = mock.MagicMock()
        mock_backend_class.get_backend_id.return_value = 'custom'
        mock_backend_instance = mock.MagicMock()
        mock_backend_class.return_value = mock_backend_instance
        registry.add(mock_backend_class)
        registry.send(
            phone_number='12345678',
            message='Testmessage',
            backend_id='custom'
        )
        mock_backend_class.assert_called_once_with(
            phone_number='12345678',
            message='Testmessage'
        )
        mock_backend_instance.send.assert_called_once_with()



from django import test

from ievv_opensource.ievv_logging.management.commands.ievv_opensource_logging_items_delete_older_than_last_100 import \
    delete_older_than_the_last_100
from ievv_opensource.ievv_logging.models import IevvLoggingEventBase, IevvLoggingEventItem
from ievv_opensource.ievv_logging.utils import IevvLogging


class TestManagmentScriptThatCleansOldLoggingItems(test.TestCase):

    def test_management_script_one_slug_only(self):
        # Make 105 log items
        for i in range(105):
            ievvlogging = IevvLogging('foo_bar')
            ievvlogging.begin()
            ievvlogging.finish()
        # check that we have in fact 105
        self.assertEqual(105, IevvLoggingEventItem.objects.count())
        # check that the last is the first created
        self.assertEqual(1, IevvLoggingEventItem.objects.last().id)
        # run the clean script
        delete_older_than_the_last_100()
        # check that count is now 100
        self.assertEqual(100, IevvLoggingEventItem.objects.count())
        # check that the last created still exists
        self.assertEqual(105, IevvLoggingEventItem.objects.first().id)
        # check that the oldest no longer exist
        with self.assertRaises(IevvLoggingEventItem.DoesNotExist):
            IevvLoggingEventItem.objects.get(pk=1)

    def test_management_script_with_multiple_slugs(self):
        # make more than 100 for multiple slugs
        sluglist = ['foo', 'bar', 'baz']
        for slug in sluglist:
            for i in range(115):
                ievvlogging = IevvLogging(slug)
                ievvlogging.begin()
                ievvlogging.finish()
        delete_older_than_the_last_100()
        # check that the number in database after cleaning is number of slugs times 100
        self.assertEqual(100*len(sluglist), IevvLoggingEventItem.objects.count())




from django import test
from django.core.management import call_command
from django.db import connection
from django.db.migrations.loader import MigrationLoader


class MigrationTestCase(test.TransactionTestCase):
    """
    Test case for a Django database migration.

    Example::

        class TestSomeMigrations(MigrationTestCase):
            migrate_from = '0002_previous_migration'
            migrate_to = '0003_migration_being_tested'

            def test_is_selected_is_flipped(self):
                MyModel = self.apps_before.get_model('myapp', 'MyModel')
                MyModel.objects.create(
                    name='Test1',
                    is_selected=True
                )
                MyModel.objects.create(
                    name='Test2',
                    is_selected=False
                )
                MyModel.objects.create(
                    name='Test3',
                    is_selected=True
                )

                self.migrate()

                MyModel = self.apps_after.get_model('myapp', 'MyModel')
                self.assertEqual(MyModel.objects.filter(is_selected=True).count, 1)
                self.assertEqual(MyModel.objects.filter(is_selected=False).count, 2)
    """

    #: The django app_label for the app you are migrating. This is the same app_label as you
    #: use with ``python manage.py makemigrations <app_label>`` to create the migration.
    app_label = None

    #: Dependencies. A list of ``(app_label, migration_name)`` tuples.
    migrate_dependencies = None

    #: Same as :obj:`~.MigrationTestCase.migrate_dependencies`, but ONLY for :obj:`~.MigrationTestCase.migrate_from`.
    migrate_from_dependencies = None

    #: Same as :obj:`~.MigrationTestCase.migrate_dependencies`, but ONLY for :obj:`~.MigrationTestCase.migrate_from`.
    migrate_to_dependencies = None

    #: The name of the migration to migrate from.
    #: Can be the full name, or just the number (I.E.: ``0002`` or ``0002_something``.
    migrate_from = None

    #: The name of the migration to migrate to.
    #: Can be the full name, or just the number (I.E.: ``0003`` or ``0003_something``.
    migrate_to = None

    @classmethod
    def _validate_dependency_list(cls, migration_list_name):
        migration_list = getattr(cls, migration_list_name)
        if not migration_list:
            return
        for app_label, migration_name in migration_list:
            if app_label == cls.app_label:
                raise ValueError(
                    'The "{}" attribute can not contain any migrations with the '
                    'same app_label as the "app_label" attribute ({})'.format(
                        app_label, cls.app_label
                    )
                )

    @classmethod
    def _validate_class_variables(cls):
        if not cls.app_label or not cls.migrate_from or not cls.migrate_to:
            raise ValueError('app_label, migrate_from and migrate_to must be specified.')
        cls._validate_dependency_list('migrate_from_dependencies')
        cls._validate_dependency_list('migrate_to_dependencies')
        cls._validate_dependency_list('migrate_dependencies')

    @classmethod
    def setUpClass(cls):
        cls._validate_class_variables()

    def _build_migration_list(self, migration_name, migrate_dependencies=None):
        migration_list = []
        if migrate_dependencies:
            migration_list.extend(migrate_dependencies)
        migration_list.append((self.app_label, migration_name))
        return migration_list

    def setUp(self):
        """
        Perform required setup.

        If you override ``setUp()``, you must call ``super().setUp()``!
        """
        self._migrate_from_list = self._build_migration_list(
            self.migrate_from,
            migrate_dependencies=self.migrate_from_dependencies or self.migrate_dependencies)
        self._migrate_to_list = self._build_migration_list(
            self.migrate_to,
            migrate_dependencies=self.migrate_to_dependencies or self.migrate_dependencies)
        self._migrate_to_list = [(self.app_label, self.migrate_to)]
        self._apps_before = self._get_apps_for_migration(self._migrate_from_list)
        self._apps_after = None
        for app_label, migration_name in self._migrate_from_list:
            self._run_migrate(app_label, migration_name)

    def _get_apps_for_migration(self, migration_states):
        loader = MigrationLoader(connection)
        full_names = []
        for app_label, migration_name in migration_states:
            if migration_name != 'zero':
                migration = loader.get_migration_by_prefix(app_label, migration_name)
                full_names.append((app_label, migration.name))
        state = loader.project_state(full_names)
        return state.apps

    @property
    def apps_before(self):
        """
        Get an ``apps`` object just like the first argument to a Django data migration
        at the state before migration has been run.

        Only available **before** :meth:`.migrate` has been called, or after :meth:`.reverse_migrate`
        has been called.
        """
        if self._apps_after is not None:
            raise AttributeError('apps_before is only available before migrate() has been run, '
                                 'or after reverse_migrate().')
        return self._apps_before

    @property
    def apps_after(self):
        """
        Get an ``apps`` object just like the first argument to a Django data migration
        at the state after migration has been run, and not available after :meth:`.reverse_migrate`
        has been called (unless :meth:`.migrate` is called again).

        Only available **after** :meth:`.migrate` has been called.
        """
        if self._apps_after is None:
            raise AttributeError('apps_after is only available after migrate() has been run. '
                                 'It is not available after reverse_migrate() unless migrate() '
                                 'has been run again.')
        return self._apps_after

    def migrate(self):
        """
        Migrate the database from :obj:`.migrate_from` to :obj:`.migrate_to`.
        """
        if self._apps_after is not None:
            raise Exception('migrate() already run. Can not run migrate() multiple times '
                            'without running reverse_migrate() in between.')

        for app_label, migration_name in self._migrate_to_list:
            self._run_migrate(app_label, migration_name)
        self._apps_after = self._get_apps_for_migration(self._migrate_to_list)

    def reverse_migrate(self):
        """
        Migrate the database from :obj:`.migrate_to` to :obj:`.migrate_from`.

        You must call :meth:`.migrate` before calling this.
        """
        if self._apps_after is None:
            raise Exception('You must run migrate() before you can run reverse_migrate().')

        for app_label, migration_name in self._migrate_from_list:
            self._run_migrate(app_label, migration_name)
        self._apps_after = None

    def get_migrate_command_kwargs(self):
        """
        Get kwargs for the ``migrate`` management command.

        The defaults are sane, by you may want to override this and change
        the ``verbosity`` argument for debugging purposes.
        """
        return {'verbosity': 0,
                'no_initial_data': True,
                'interactive': False}

    def _run_migrate(self, app_label, migration_name, fake=False):
        kwargs = self.get_migrate_command_kwargs()
        kwargs['fake'] = fake
        args = ('migrate', app_label, migration_name)
        call_command(*args, **kwargs)



from ievv_opensource.ievv_sms import sms_registry
from ievv_opensource.ievv_sms.models import DebugSmsMessage


class Backend(sms_registry.AbstractSmsBackend):
    """
    Backend that creates DebugSmsMessage object

    Useful for development.

    The backend_id for this backend is ``debug_dbstore``.
    """
    @classmethod
    def get_backend_id(cls):
        return 'debug_dbstore'

    def send(self):

        db_record = DebugSmsMessage()
        db_record.phone_number = self.phone_number
        db_record.message = self.message
        db_record.send_as = self.send_as or ''
        db_record.full_clean()
        db_record.save()



import threading
import time
import os

import psutil

from ievv_opensource.utils import logmixin
from ievv_opensource.utils import shellcommandmixin


class AbstractRunnableThread(threading.Thread, logmixin.LogMixin):
    """
    Abstract class for a thread that runs something for the ``ievvdevrun`` framework.

    You have to override the ``run()``-method (refer to the docs for ``threading.Thread.run``),
    and the :meth:`~AbstractRunnableThread.stop` method.
    """
    def __init__(self, name=None):
        """
        Parameters:
            name: An optional name for the logger.
                See :meth:`~AbstractRunnableThread.get_logger_name`.
        """
        self.logger_name = name
        super(AbstractRunnableThread, self).__init__()

    def log_startup(self):
        """
        Called automatically on startup with ``"Starting <name>"`` message.
        """
        self.get_logger().info('Starting {}'.format(self.get_logger_name()))

    def log_successful_stop(self, message=''):
        """
        Call this in :meth:`.stop` to show a message after successfully stopping
        the runnable.

        Parameters:
            message: Optional extra message to show.
        """
        self.get_logger().info('{} stopped ! {}'.format(
            self.get_logger_name(), message))

    def start(self):
        self.log_startup()
        return super(AbstractRunnableThread, self).start()

    def get_logger_name(self):
        """
        Get the logger name. Defaults to the ``name``-parameter and falls back
        on the module name of the class.
        """
        return self.logger_name or self.__class__.__module__

    def stop(self):
        """
        Must stop the code running in the thread (the code in the run method).
        """
        raise NotImplementedError()


class ShellCommandRunnableThread(AbstractRunnableThread,
                                 shellcommandmixin.ShellCommandMixin):
    """
    Makes it very easy to implement :class:`.AbstractRunnableThread`
    for system/shell commands.

    Examples:

        Implement Django runserver::

            class DjangoRunserverRunnableThread(base.ShellCommandRunnableThread):
                def get_logger_name(self):
                    return 'Django development server'

                def get_command_config(self):
                    return {
                        # We use sys.executable instead of "python" to ensure we run the same python executable
                        # as we are using to start the process.
                        'executable': sys.executable,
                        'args': ['manage.py', 'runserver']
                    }

        As you can see, we just need specify the name and the command we want
        to run. You can even do this without creating a class by
        sending the command config as a parameter to this class
        in your django settings::

            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': ievvdevrun.config.RunnableThreadList(
                    ievvdevrun.runnables.base.ShellCommandRunnableThread(
                        name='Django development server',
                        command_config={
                            'executable': sys.executable,
                            'args': ['manage.py', 'runserver']
                        }
                    ),
                )
            }


        Making the command automatically restart when it crashes::

            class DjangoRunserverRunnableThread(base.ShellCommandRunnableThread):
                default_autorestart_on_crash = True
                # ... same code as the example above ...

        You can also autorestart by sending ``autorestart_on_crash=True``
        as a parameter for the class.
    """

    #: The default value for the ``autorestart_on_crash`` parameter.
    #: You can override this in subclasses if it is more natural
    #: to automatically restart on crash by default.
    default_autorestart_on_crash = False

    def __init__(self, name=None, command_config=None, autorestart_on_crash=None):
        """
        Parameters:
            name: Optional name for the runnable. Defaults to the module name.
            command_config: Same format as the return value from
                :meth:`.get_command_config`.
            autorestart_on_crash: Set this to ``True`` if you want to
                automatically restart the process if it crashes.
        """
        self.command_config = command_config
        if autorestart_on_crash is None:
            autorestart_on_crash = self.default_autorestart_on_crash
        self.autorestart_on_crash = autorestart_on_crash
        super(ShellCommandRunnableThread, self).__init__(name=name)

    def _restartloop(self):
        while True:
            if self.is_running:
                if self.command.process.is_alive():
                    try:
                        process = psutil.Process(self.command.pid)
                    except psutil.NoSuchProcess:
                        pass
                    else:
                        self.detected_process_ids.add(self.command.pid)
                        for childprocess in process.children():
                            self.detected_process_ids.add(childprocess.pid)
                else:
                    self.get_logger().warning('Restarting "{}" because of crash.'.format(
                        self.get_logger_name()))
                    self._start_command(restart=True)
            else:
                return

            # Check if alive every 5 second, but check if we have been stopped
            # every 200ms.
            for x in range(25):
                time.sleep(0.2)
                if not self.is_running:
                    break

    def _start_command(self, restart=False):
        if restart:
            self.stop()
        self.is_running = True
        self.detected_process_ids = set()
        commandconfig = self.get_command_config()
        kwargs = commandconfig.get('kwargs', {}).copy()
        kwargs.update({'_bg': True})
        environment = commandconfig.get('environment')
        _env = None
        if environment:
            _env = os.environ.copy()
            _env.update(environment)
        self.command = self.run_shell_command(commandconfig['executable'],
                                              args=commandconfig.get('args', []),
                                              kwargs=kwargs,
                                              _env=_env)

    def get_command_config(self):
        """
        Get the config for the shell/system command as a dict with the
        following keys:

        - ``executable``: The name of the executable (E.g.: ``python``, ``sqlite``, ...).
        - ``args``: Arguments for the executable as a list.
        - ``kwargs``: Keyword arguments for the executable as a dict. ``my_option: "myvalue"``
            is automatically translated to ``--my-option="myvalue"``.
        """
        if self.command_config:
            return self.command_config
        else:
            raise NotImplementedError('You must override get_command_config() or send '
                                      'command_config to __init__().')

    def run(self):
        self._start_command()
        if self.autorestart_on_crash:
            self._restartloop()

    def stop(self):
        self.is_running = False
        process_ids = set(self.terminate_process(self.command.pid))
        for process_id in self.detected_process_ids:
            process_ids.update(set(self.terminate_process(process_id)))
        self.log_successful_stop(
            '[killed pids: {}]'.format(', '.join(map(str, process_ids))))



import json

from ievv_opensource.utils.ievvbuildstatic import pluginbase
from ievv_opensource.utils.shellcommandmixin import ShellCommandError
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin


class Plugin(pluginbase.Plugin, ShellCommandMixin):
    """
    Bower install plugin --- installs bower packages.

    The packages are installed when ``ievv buildstatic`` starts up.
    The plugin creates a ``bower.json`` file, and runs ``bower install``
    using the created ``bower.json``-file.

    You will most likely want to add ``bower.json`` and ``bower_components``
    to your VCS ignore file.

    Examples:

        Install bootstrap 3.1.1 and angularjs 1.4.1::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.bowerinstall.Plugin(
                            packages={
                                'bootstrap': '~3.1.1',
                                'angular': '~1.4.1'
                            }
                        ),
                    ]
                )
            )
    """
    name = 'bowerinstall'

    def __init__(self, packages, **kwargs):
        super(Plugin, self).__init__(**kwargs)
        self.packages = packages

    def get_bowerjson_path(self):
        return self.app.get_source_path('bower.json')

    def create_bowerjson(self):
        packagedata = {
            'name': self.app.appname,
            'version': self.app.version,
            "private": True,
            "interactive": False,
            "analytics": False,
            "devDependencies": self.packages
        }
        open(self.get_bowerjson_path(), 'wb').write(
            json.dumps(packagedata, indent=2).encode('utf-8'))

    def get_bower_version(self):
        return None

    def install(self):
        self.app.get_installer('npm').queue_install(
            'bower', version=self.get_bower_version())

    def run(self):
        self.get_logger().command_start('Running bower install for {}'.format(
            self.app.get_source_path()))
        self.create_bowerjson()
        bower_executable = self.app.get_installer('npm').find_executable('bower')
        try:
            self.run_shell_command(bower_executable,
                                   args=['install'],
                                   _cwd=self.app.get_source_path())
        except ShellCommandError:
            self.get_logger().command_error('bower install FAILED!')
            raise SystemExit()
        else:
            self.get_logger().command_success('bower install succeeded :)')



import os
import tempfile
import shutil

from django.core import management
from django.test import TestCase


class TestDumpDbAsJson(TestCase):
    def __remove_dumpdata_directory(self):
        if os.path.exists(self.dumpdata_directory):
            shutil.rmtree(self.dumpdata_directory)

    def setUp(self):
        self.dumpdata_directory = tempfile.mkdtemp()
        self.dumpdata_file = os.path.join(self.dumpdata_directory, 'default.sql')
        self.__remove_dumpdata_directory()

    def tearDown(self):
        self.__remove_dumpdata_directory()

    def test_dumpfile_created(self):
        self.assertFalse(os.path.exists(self.dumpdata_file))
        with self.settings(IEVVTASKS_DUMPDATA_DIRECTORY=self.dumpdata_directory):
            management.call_command('ievvtasks_dump_db_as_sql')
        self.assertTrue(os.path.exists(self.dumpdata_file))



from django.core import management
from django.core.management import BaseCommand


class BaseIevvTasksCommand(BaseCommand):
    def run_management_commands(self, management_commands):
        for management_command in management_commands:
            if isinstance(management_command, dict):
                args = management_command.get('args', [])
                options = management_command.get('options', {})
                management_command = management_command['name']
            else:
                args = []
                options = {}
            self.stdout.write(
                'Running management command: {} with '
                'args={!r} and options={!r}.'.format(
                    management_command, args, options))
            management.call_command(management_command, *args, **options)



import json
import os

from ievv_opensource.utils.ievvbuildstatic import pluginbase
from .gzip_compress_mixin import GzipCompressMixin
# from ievv_opensource.utils.ievvbuildstatic.installers.npm import NpmInstaller
from ievv_opensource.utils.shellcommandmixin import (ShellCommandError,
                                                     ShellCommandMixin)


class CssBuildException(Exception):
    """
    Raised when :meth:`.AbstractPlugin.build_css` fails.
    """


class AbstractPlugin(pluginbase.Plugin, ShellCommandMixin, GzipCompressMixin):
    """
    Base class for builders that produce CSS.
    """
    default_group = 'css'

    def __init__(self, lint=True, lintrules=None, lintrules_overrides=None, autoprefix=True,
                 browserslist='> 5%', gzip=False, gzip_compresslevel=9, **kwargs):
        """

        Args:
            lint (bool): Lint the styles? Defaults to ``True``.
            lintrules (dict): Rules for the linter. See https://github.com/stylelint/stylelint.
                Defaults to::

                    {
                        "block-no-empty": None,
                        "color-no-invalid-hex": True,
                        "comment-empty-line-before": ["always", {
                            "ignore": ["stylelint-commands", "after-comment"],
                        }],
                        "declaration-colon-space-after": "always",
                        "indentation": 4,
                        "max-empty-lines": 2
                    }

            lintrules_overrides (dict): Overrides for ``lintrules``. Use this
                if you just want to add new rules or override some of the existing rules.
            autoprefix (bool): Run https://github.com/postcss/autoprefixer on the css?
                Defaults to ``True``.
            browserslist (str): A string with defining supported browsers
                for https://github.com/ai/browserslist. Used by the autoprefix
                and cssnano commands.
            gzip (bool): Make a .css.gz version of the file in --production mode. This is added in
                addition to the .css file (same filename, just with .gz at the end).
            gzip_compresslevel (int, optional): Gzip compression level.
                A number between 0 and 9. Higher is better compression,
                but slower to compress. Defaults to 9.
            **kwargs: Kwargs for :class:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin`.
        """
        super(AbstractPlugin, self).__init__(**kwargs)
        self.lint = lint
        self.autoprefix = autoprefix
        self.browserslist = browserslist
        self.lintrules = lintrules or {
            "block-no-empty": None,
            "color-no-invalid-hex": True,
            "comment-empty-line-before": ["always", {
                "ignore": ["stylelint-commands", "after-comment"],
            }],
            "declaration-colon-space-after": "always",
            "indentation": 4,
            "max-empty-lines": 2
        }
        if lintrules_overrides:
            self.lintrules.update(lintrules_overrides)
        self.gzip = gzip
        self.gzip_compresslevel = gzip_compresslevel

    def gzip_compression_enabled(self):
        return self.gzip and self.app.apps.is_in_production_mode()

    def gzip_compresslevel(self):
        return self.gzip_compresslevel

    def get_filepaths_to_gzip(self):
        return [
            self.get_destinationfile_path()
        ]

    def get_postcss_cli_version(self):
        return None

    def get_autoprefixer_version(self):
        return None

    def get_cssnano_version(self):
        return None

    def get_stylelint_version(self):
        return None

    def should_minify(self):
        return self.app.apps.is_in_production_mode()

    def requires_postcss(self):
        return self.autoprefix or self.should_minify()

    def install(self):
        if self.requires_postcss():
            self.app.get_installer('npm').queue_install(
                'postcss-cli', version=self.get_postcss_cli_version())
            if self.autoprefix:
                self.app.get_installer('npm').queue_install(
                    'autoprefixer', version=self.get_autoprefixer_version())
            if self.should_minify():
                self.app.get_installer('npm').queue_install(
                    'cssnano', version=self.get_cssnano_version())
        if self.lint:
            self.app.get_installer('npm').queue_install(
                'stylelint', version=self.get_stylelint_version())

    def build_css(self, temporary_directory):
        """
        Override this method and implement the code to build the css.
        """
        raise NotImplementedError()

    def get_destinationfile_path(self):
        """
        Override this method and return the absolute path of the destination/output css
        file.
        """
        raise NotImplementedError()

    def run_postcss(self):
        args = []
        if self.autoprefix:
            args.extend([
                '--use', 'autoprefixer',
                '--autoprefixer.browsers', self.browserslist,
            ])
        if self.should_minify():
            args.extend([
                '--use', 'cssnano',
                '--cssnano.browsers', self.browserslist,
                '--no-map'
            ])
        self.get_logger().command_start('Running postcss with [{args}] on {destination}'.format(
            args=' '.join(args),
            destination=self.get_destinationfile_path()))
        args.extend([
            '--output', self.get_destinationfile_path(),
            self.get_destinationfile_path()
        ])
        try:
            self.run_shell_command(
                self.app.get_installer('npm').find_executable('postcss'),
                args=args)
        except ShellCommandError:
            self.get_logger().error('postcss failed!')
            raise CssBuildException()
        else:
            self.get_logger().success('postcss ran successfully :)')

    def get_all_source_file_paths(self):
        """
        Used for utilities like the linter to get a list of all
        source files.

        You must override this in subclasses.
        """
        raise NotImplementedError()

    def make_lintconfig_file(self, temporary_directory):
        lintconfig_path = os.path.join(temporary_directory, 'stylelint.json')
        lintconfig = {
            'rules': self.lintrules,
        }
        open(lintconfig_path, 'wb').write(json.dumps(lintconfig, indent=2).encode('utf-8'))
        return lintconfig_path

    def lint_styles(self, temporary_directory):
        lintconfig_path = self.make_lintconfig_file(temporary_directory=temporary_directory)
        self.get_logger().command_start('Linting styles')
        args = [
            self.get_all_source_file_paths()[0],
            '--config', lintconfig_path
        ]
        try:
            self.run_shell_command(
                self.app.get_installer('npm').find_executable('stylelint'),
                args=args)
        except ShellCommandError:
            self.get_logger().error('Stylesheet linting failed!')
        else:
            self.get_logger().success('Stylesheet linting was successful :)')

    def preprocess_styles(self, temporary_directory):
        if self.lint:
            self.lint_styles(temporary_directory=temporary_directory)

    def postprocess_styles(self, temporary_directory):
        if self.requires_postcss():
            self.run_postcss()

    def run(self):
        temporary_directory = self.make_temporary_build_directory()
        try:
            self.preprocess_styles(temporary_directory=temporary_directory)
            self.build_css(temporary_directory=temporary_directory)
            self.postprocess_styles(temporary_directory=temporary_directory)
        except CssBuildException:
            pass
        else:
            self.gzip_compress_files()



from django.apps import AppConfig


class ModelSmsAppConfig(AppConfig):
    name = 'ievv_opensource.ievv_sms'
    verbose_name = "IEVV SMS"

    def ready(self):
        from ievv_opensource.ievv_sms import sms_registry
        from ievv_opensource.ievv_sms.backends import debugprint
        from ievv_opensource.ievv_sms.backends import pswin
        from ievv_opensource.ievv_sms.backends import debug_dbstore
        registry = sms_registry.Registry.get_instance()
        registry.add(debugprint.Backend)
        registry.add(debugprint.Latin1Backend)
        registry.add(pswin.Backend)
        registry.add(debug_dbstore.Backend)



import json
import os

from ievv_opensource.utils.ievvbuildstatic import pluginbase
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin


class Plugin(pluginbase.Plugin, ShellCommandMixin):
    """
    Autosetup esdoc config and npm script.

    The plugin creates the esdoc.json config file automatically.
    It also creates a directory named ``esdoc`` and a file
    named ``index.md`` within that directory. That file is
    the frontpage of your docs, and you should edit it and provide
    some information about the app/library.

    If you want to add a manual "page" to your docs, you do that
    by adding files and/or directories to the ``esdoc/manual/`` directory
    (relative to the source directory - the directory with package.json).
    Esdoc has a specific set of manual sections that you can add:

    - asset
    - overview
    - installation
    - usage
    - tutorial
    - configuration
    - example
    - faq
    - changelog

    This plugin automatically adds these to the ``esdoc.json`` with the following
    rules:

    - Does ``esdoc/manual/<section>.md`` exist? Add that as the first entry in
      the ``manual.<section>`` list.
    - Does ``esdoc/manual/<section>/`` exist? Add all .md files in that directory
      to the ``manual.<section>`` list in sorted order.

    Examples:

        It is really simple to use::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.autosetup_esdoc.Plugin(),
                    ]
                )
            )

        Lets add a tutorial! First create ``esdoc/manual/tutorial.md`` and run ievv buildstatic.
        That file should end up in ``esdoc.json`` and if you build the docs
        you should get a "manual" link at the top of the manual. If your tutorial grows,
        simply create the ``esdoc/manual/tutorial/`` directory and add markdown files to
        it. The files are sorted by filename and the output in the docs is still
        a single file. The ``esdoc/manual/tutorial.md`` file will still end up first if
        it is present. You can choose if you want to use the ``esdoc/manual/tutorial.md``,
        the ``esdoc/manual/tutorial/`` directory or both. The tutorial example works
        for all the manual sections documented above.
    """
    name = 'autosetup_esdoc'

    def __init__(self, title=None, **kwargs):
        """

        Args:
            title (str): The title of the docs. Falls back to the app name
                if not specified.
            **kwargs: Kwargs for :class:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin`.
        """
        super(Plugin, self).__init__(**kwargs)
        self.title = title

    def install(self):
        """
        Installs the ``esdoc`` and ``esdoc-importpath-plugin``
        npm packages as dev dependencies.
        """
        self.app.get_installer('npm').queue_install(
            package='esdoc',
            installtype='dev'
        )
        self.app.get_installer('npm').queue_install(
            package='esdoc-importpath-plugin',
            installtype='dev'
        )

    def get_esdoc_config_path(self):
        return self.app.get_source_path('esdoc.json')

    def get_manual_path(self, *path):
        return self.app.get_source_path('esdoc', 'manual', *path)

    def get_esdoc_index_path(self):
        return self.app.get_source_path('esdoc', 'index.md')

    def esdoc_index_exists(self):
        return os.path.exists(self.get_esdoc_index_path())

    def get_default_esdoc_index_content(self):
        return '**TODO:** edit this text in ``{esdoc_index_path}``.'.format(
            esdoc_index_path=self.app.make_source_relative_path(
                self.get_esdoc_index_path())
        )

    def create_default_esdoc_index(self):
        filepath = self.get_esdoc_index_path()
        directorypath = os.path.dirname(filepath)
        if not os.path.exists(directorypath):
            os.makedirs(directorypath)
        open(filepath, 'w').write(
            self.get_default_esdoc_index_content())

    def _add_manual_dict_files_for_section(self, section):
        directory_path = self.get_manual_path(section)
        file_path = self.get_manual_path('{}.md'.format(section))
        file_list = []
        if os.path.exists(file_path):
            file_list.append(self.app.make_source_relative_path(file_path))
        if os.path.exists(directory_path):
            for filename in sorted(os.listdir(directory_path)):
                if filename.endswith('.md'):
                    path = self.app.make_source_relative_path(
                        os.path.join(directory_path, filename))
                    file_list.append(path)
        return file_list

    def make_manual_dict(self):
        manual_dict = {}
        for section in ('asset', 'overview', 'installation', 'usage',
                        'tutorial', 'configuration', 'example', 'faq',
                        'changelog'):
            file_list = self._add_manual_dict_files_for_section(section=section)
            if file_list:
                manual_dict[section] = file_list
        return manual_dict

    def make_plugins_list(self):
        return [
            {
                "name": "esdoc-importpath-plugin",
                "option": {
                    "replaces": [
                        {
                            "from": "^javascript/",
                            "to": ""
                        },
                        {
                            "from": "^{appname}/".format(appname=self.app.appname),
                            "to": ""
                        }
                    ]
                }
            }
        ]

    def make_esdoc_config_dict(self):
        config_dict = {
            "title": self.title or self.app.appname,
            "source": self.app.make_source_relative_path("scripts", "javascript"),
            "destination": self.app.make_source_relative_path("built_docs"),
            "includes": ["\\.(js|es6|jsx)$"],
            "excludes": [
                "__tests__.+$",
                "__mocks__.+$"
            ],
            "index": self.app.make_source_relative_path(self.get_esdoc_index_path()),
        }
        plugins_list = self.make_plugins_list()
        if plugins_list:
            config_dict['plugins'] = plugins_list
        manual_dict = self.make_manual_dict()
        if manual_dict:
            config_dict['manual'] = manual_dict
        return config_dict

    def create_esdoc_config(self):
        open(self.get_esdoc_config_path(), 'w').write(
            json.dumps(
                self.make_esdoc_config_dict(),
                indent=2,
                sort_keys=True
            ))

    def run(self):
        self.get_logger().command_start('Autosetup esdoc for {appname}'.format(
            appname=self.app.appname))
        if not self.esdoc_index_exists():
            self.create_default_esdoc_index()
        self.create_esdoc_config()
        self.app.get_installer('npm').add_npm_script(
            'build-docs', 'esdoc -c {configpath}'.format(
                configpath=self.app.make_source_relative_path(self.get_esdoc_config_path())))
        self.get_logger().command_success('Autoset esdoc config succeeded :)')



import datetime
import decimal
import json

import arrow
from django.db import models


class Encoder(json.JSONEncoder):
    """
    Extends the default json.JSONEncoder with support for more types.

    Adds support for:

    - `datetime.datetime`
        - Serialized as ``{"__type__": "datetime", "value": "<iso datetime with timezone>"}``
    - `datetime.date`
        - Serialized as ``{"__type__": "date", "value": "<iso date YYYY-MM-DD>"}``
    - `datetime.time`
        - Serialized as ``{"__type__": "time", "value": "<str(timeobject)>"}``
    - `decimal.Decimal`
        - Serialized as ``{"__type__": "decimal", "value": "<str(decimalvalue)>"}``
    - Django model objects
        - Serialized as ``{"__type__": "djangomodel", "__model__": "someapp.SomeModel", "value": "<model object PK>"}``
    """
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return {
                '__type__': 'datetime',
                'value': arrow.get(obj).format()
            }
        elif isinstance(obj, datetime.date):
            return {
                '__type__': 'date',
                'value': obj.isoformat()
            }
        elif isinstance(obj, datetime.time):
            return {
                '__type__': 'time',
                'value': obj.isoformat()
            }
        elif isinstance(obj, decimal.Decimal):
            return {
                '__type__': 'decimal',
                'value': str(obj)
            }
        elif isinstance(obj, models.Model):
            return {
                '__type__': 'djangomodel',
                '__model__': f'{obj._meta.label}',
                'value': obj.pk
            }
        return super().default(obj)


class Decoder(json.JSONDecoder):
    """
    Extends the default json.JSONDecoder with support for the types that :class:`.Encoder` supports.

    Note that even though :class:`.Encoder` supports Django models, we do not support
    looking up the actual database object. So decoding Django models just returns
    them with the same representation that they where encoded with (model + PK).
    """
    def __init__(self):
        super().__init__(object_hook=self.custom_object_hook)

    def custom_object_hook(self, d):
        if '__type__' not in d or 'value' not in d:
            return d
        _type = d['__type__']
        value = d['value']
        if _type == 'datetime':
            return arrow.get(value).datetime
        elif _type == 'date':
            return arrow.get(value).datetime.date()
        elif _type == 'time':
            return arrow.get(f'2000-01-01 {value}').datetime.time()
        elif _type == 'decimal':
            return decimal.Decimal(value)
        # NOTE: We do not support reverse lookup of django models in decoder!
        return d


def dumps(data, indent=None, sort_keys=False):
    """
    For most use-cases this is a drop in replacement for json.dumps()
    that just extends the supported datatypes with what :class:`.Encoder` supports.
    """
    return Encoder(indent=indent, sort_keys=sort_keys).encode(data)


def loads(data):
    """
    For most use-cases this is a drop in replacement for json.loads()
    that just extends the supported datatypes with what :class:`.Decoder` supports.
    """
    return Decoder().decode(data)



import json
import os
from collections import OrderedDict

import shutil

from ievv_opensource.utils.ievvbuildstatic.installers.base import AbstractInstaller
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin


class NpmInstallerError(Exception):
    pass


class PackageJsonDoesNotExist(NpmInstallerError):
    pass


class AbstractNpmInstaller(AbstractInstaller, ShellCommandMixin):
    """
    Abstract npm installer.

    Base class for npm and yarn installers.
    """
    @classmethod
    def add_cli_arguments(cls, parser):
        parser.add_argument('--{}-clean-node-modules'.format(cls.optionprefix),
                            dest='clean_node_modules',
                            action='store_true', required=False, default=False,
                            help='Remove node_modules before installing node packages.')
        parser.add_argument('--{}-link'.format(cls.optionprefix),
                            dest='npmlink',
                            action='append',
                            metavar='PACKAGENAME',
                            required=False,
                            help='Link specified javascript package. Can be repeated multiple times.')

    def __init__(self, *args, **kwargs):
        super(AbstractNpmInstaller, self).__init__(*args, **kwargs)
        self.queued_packages = OrderedDict()

    def __validate_installtype(self, installtype):
        installtypes = (None, "dev", "optional")
        if installtype not in installtypes:
            raise ValueError(
                'Invalid installtype {installtype!r}. Must be '
                'one of {installtypes}'.format(
                    installtype=installtype,
                    installtypes=', '.join(installtypes)
                ))

    def queue_install(self, package, version=None, installtype=None):
        """
        Installs the given npm package in the build
        directory for the app.

        Does nothing if the package is already installed.

        Parameters:
            package: The package name.
            version: The package version.
            installtype: Determines where the package ends up in
                package.json. Can be one of:

                - ``None`` (the default): Ends up in dependencies.
                - ``"dev"``: Ends up in devDependencies.
                - ``"optional"`` (the default): Ends up in optionalDependencies.
        """
        self.__validate_installtype(installtype=installtype)
        self.get_logger().debug('Queued install of {} (version={}) for {}.'.format(
            package, version, self.app.appname))
        queue = True
        if package in self.queued_packages:
            queued_version = self.queued_packages[package]
            if version is None:
                queue = False
            elif queued_version is None:
                queue = True
            else:
                self.get_logger().warning(
                    'Multiple explicit versions of the same NPM package {package} '
                    'specified for {appname}: {version!r} and {queued_version!r}. '
                    'Using {version!r}.'.format(
                        package=package,
                        version=version,
                        queued_version=queued_version,
                        appname=self.app.appname))
                queue = True
        if queue:
            self.queued_packages[package] = {
                'version': version,
                'installtype': installtype
            }

    def get_packagejson_path(self):
        return self.app.get_source_path('package.json')

    def packagejson_exists(self):
        return os.path.exists(self.get_packagejson_path())

    # def packagejson_created_by_ievv_buildstatic(self):
    #     return 'ievv_buildstatic' in self.get_packagejson_dict()

    def __get_packagejson_dict(self):
        package_json_string = open(self.get_packagejson_path()).read()
        package_json_dict = json.loads(package_json_string)
        return package_json_dict

    def get_packagejson_dict(self):
        """
        Get ``package.json`` as a dict.

        This is cached, so calling it many times only requires
        one read from the disk.
        """
        if not hasattr(self, '_packagejson_dict'):
            self._packagejson_dict = self.__get_packagejson_dict()
        return self._packagejson_dict

    def save_packagejson_dict(self):
        """
        Save the dict returned by :meth:`.get_packagejson_dict`
        to ``package.json``.

        You must call this if you change the dict returned
        by :meth:`.get_packagejson_dict`.
        """
        open(self.get_packagejson_path(), 'wb').write(
            json.dumps(
                self._packagejson_dict,
                indent=2,
                sort_keys=True
            ).encode('utf-8'))

    def create_packagejson(self):
        """
        Create initial ``package.json``.
        """
        package_json_dict = {
            'name': self.app.appname,
            'private': True,
            'ievv_buildstatic': {},
            # We do not care about the version. We are not building a distributable package.
            'version': '0.0.1',
        }
        self._packagejson_dict = package_json_dict
        self.save_packagejson_dict()

    def add_npm_script(self, scriptname, script):
        """
        Add a script to the ``"scripts"`` property of ``package.json``.

        Overwrites any existing script named ``scriptname``.

        Args:
            scriptname: The key in the ``"scripts"`` object/dict.
            script: The value in in the ``"scripts"`` object/dict.
        """
        package_json_dict = self.get_packagejson_dict()
        scripts = self.get_packagejson_dict().get('scripts', {})
        scripts[scriptname] = script
        package_json_dict['scripts'] = scripts
        self.save_packagejson_dict()

    def has_npm_script(self, scriptname):
        return scriptname in self.get_packagejson_dict().get('scripts', {})

    def get_packagejson_key_from_installtype(self, installtype):
        if installtype is None:
            return 'dependencies'
        elif installtype == 'dev':
            return 'devDependencies'
        elif installtype == 'optional':
            return 'optionalDependencies'

    def package_is_in_packages_json(self, package, properties):
        package_json_dict = self.get_packagejson_dict()
        key = self.get_packagejson_key_from_installtype(installtype=properties['installtype'])
        packages = package_json_dict.get(key, {})
        return package in packages

    def get_node_modules_directory(self):
        return self.app.get_source_path('node_modules')

    def find_executable(self, executablename):
        """
        Find an executable named ``executablename``.

        Returns the absolute path to the executable.
        """
        executablepath = self.app.get_source_path(
            'node_modules', '.bin', executablename)
        return executablepath

    def get_package_json_dict_for_package(self, package_name):
        package_json_path = self.app.get_source_path(
            'node_modules', package_name, 'package.json')
        package_json_string = open(package_json_path).read()
        package_json_dict = json.loads(package_json_string)
        return package_json_dict

    def get_package_version(self, package_name):
        return self.get_package_json_dict_for_package(package_name)['version']

    def get_installed_package_names(self):
        if not os.path.exists(self.get_packagejson_path()):
            raise PackageJsonDoesNotExist()
        package_json_string = open(self.get_packagejson_path()).read()
        package_json_dict = json.loads(package_json_string)
        package_names = []
        for package_name, version in package_json_dict.get('devDependencies', {}).items():
            package_names.append(package_name)
        return package_names

    def install_queued_packages(self):
        for package, properties in self.queued_packages.items():
            if not self.package_is_in_packages_json(package=package,
                                                    properties=properties):
                self.install_npm_package(package=package, properties=properties)

    def install(self):
        raise NotImplementedError()

    def install_packages_from_packagejson(self):
        raise NotImplementedError()

    def install_npm_package(self, package, properties):
        raise NotImplementedError()

    def uninstall_npm_package(self, packagename):
        raise NotImplementedError()

    def run_packagejson_script(self, script, args=None):
        """
        Run a script in the scripts section of the package.json.

        Args:
            script: The npm script to run.
            args (list): List of arguments.
        """
        raise NotImplementedError()

    def initialize(self):
        if os.path.exists(self.get_packagejson_path()):
            self.install()

    def remove_node_modules_directory(self):
        """
        Remove node_modules directory.
        """
        if os.path.exists(self.get_node_modules_directory()):
            shutil.rmtree(self.get_node_modules_directory())

    def get_linked_packages(self):
        """
        Get all the linked packages in the node_modules directory.
        (packages added with npm link).

        """
        node_modules_directory = self.get_node_modules_directory()
        linked_packages = set()
        if os.path.exists(node_modules_directory):
            for filename in os.listdir(node_modules_directory):
                path = os.path.join(node_modules_directory, filename)
                if os.path.islink(path):
                    linked_packages.add(filename)
        return linked_packages

    def _remove_linked_packages(self, linked_packages):
        for packagename in linked_packages:
            self._unlink_package(packagename=packagename)
            path = os.path.join(self.get_node_modules_directory(), packagename)
            if os.path.exists(path):
                os.remove(path)

    def _link_package(self, packagename):
        self.get_logger().info('Linking package {!r}'.format(packagename))
        self.link_package(packagename=packagename)
        self.add_deferred_success(
            'linked package {}'.format(packagename)
        )

    def link_package(self, packagename):
        raise NotImplementedError()

    def _unlink_package(self, packagename):
        self.get_logger().info('Unlinking package {!r}'.format(packagename))
        self.unlink_package(packagename=packagename)
        self.add_deferred_warning(
            'unlinked package {}'.format(packagename)
        )

    def unlink_package(self, packagename):
        raise NotImplementedError()

    def get_requested_link_packages_from_options(self):
        return self.get_option('npmlink', []) or []

    def handle_linked_packages(self):
        linked_packages = self.get_linked_packages()
        if self.is_in_production_mode() and linked_packages:
            self.get_logger().warning(
                'There are linked packages ({linked_packages!r}) in {node_modules_directory!r}. '
                'Since we are running in production mode, we remove the node_modules directory '
                'to make sure we get a prestine build.'.format(
                    linked_packages=linked_packages,
                    node_modules_directory=self.get_node_modules_directory()
                ))
            self.remove_node_modules_directory()
        else:
            requested_link_packages = set(self.get_requested_link_packages_from_options())
            packages_to_unlink = linked_packages.difference(requested_link_packages)
            packages_to_link = requested_link_packages.difference(linked_packages)
            if packages_to_unlink:
                self.get_logger().warning(
                    'There are linked packages ({linked_packages!r}) in {node_modules_directory!r} '
                    'that has not been specified using command line options. '
                    'Unlinking these packages.'.format(
                        linked_packages=packages_to_unlink,
                        node_modules_directory=self.get_node_modules_directory()
                    ))
                self._remove_linked_packages(
                    linked_packages=packages_to_unlink)
            for packagename in packages_to_link:
                self._link_package(packagename=packagename)
            if linked_packages:
                self.add_deferred_warning('Installing with linked packages: {!r}'.format(linked_packages))



import sys

from ievv_opensource.utils.ievvdevrun.runnables import base


class RunnableThread(base.ShellCommandRunnableThread):
    """
    Django runserver runnable thread.

    Examples:

        You can just add it to your Django development settings with::

            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': ievvdevrun.config.RunnableThreadList(
                    ievvdevrun.runnables.django_runserver.RunnableThread()
                )
            }

        And you can make it not restart on crash with::

            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': ievvdevrun.config.RunnableThreadList(
                    ievvdevrun.runnables.django_runserver.RunnableThread(
                        autorestart_on_crash=False)
                )
            }

    """
    default_autorestart_on_crash = True

    def __init__(self, host='127.0.0.1', port='8000', environment=None, **kwargs):
        """
        Args:
            host: The host to run the Django server on. Defaults to ``"127.0.0.1"``.
            port: The port to run the Django server on. Defaults to ``"8000"``.
            **kwargs: Sent to manage.py runserver as --<key> <value> arguments.
                If the value is True, we only send --<key>.
        """
        self.host = host
        self.port = port
        self.kwargs = kwargs
        self.environment = environment
        super(RunnableThread, self).__init__()

    def get_logger_name(self):
        return 'Django development server'

    def get_command_config(self):
        args = ['manage.py', 'runserver', '{}:{}'.format(self.host, self.port)]
        for key, value in self.kwargs.items():
            args.append(f'--{key}')
            if value != True:
                args.append(str(value))
        return {
            'executable': sys.executable,
            'args': args,
            'environment': self.environment
        }



import sys
from ievv_opensource.utils.ievvdevrun.runnables import base


class RunnableThread(base.ShellCommandRunnableThread):
    """
    Django runserver runnable thread.
    Examples:
        You can just add it to your Django development settings with::
            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': devrun_rq_runnable.RunnableThreadList(
                    ievvdevrun.runnables.django_runserver.RunnableThread()
                )
            }
        And you can make it not restart on crash with::
            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': devrun_rq_runnable.RunnableThreadList(
                    ievvdevrun.runnables.django_runserver.RunnableThread(
                        autorestart_on_crash=False)
                )
            }
    """
    default_autorestart_on_crash = True

    def __init__(self, queuename='default'):
        """
        Args:
            queuename: The name of the queue
        """
        self.queuename = queuename
        super(RunnableThread, self).__init__()

    def get_logger_name(self):
        return 'RQ queue worker'

    def get_command_config(self):
        return {
            'executable': sys.executable,
            'args': ['manage.py', 'rqworker', self.queuename]
        }



from django.conf import settings
from django.urls import NoReverseMatch, Resolver404, resolve, reverse
from django.utils import translation
from django.utils.translation import get_language_info
from ievv_opensource.ievv_i18n_url import active_i18n_url_translation
from ievv_opensource.ievv_i18n_url.base_url import BaseUrl


class UrlTransformError(Exception):
    """
    Raised when :meth:`.AbstractHandler.transform_url_to_languagecode`
    or :meth:`.AbstractHandler.transform_urlpath_to_languagecode` fails
    to transform the URL.
    """


class AbstractHandler:
    """
    Base class for `ievv_i18n_url` handlers.
    """
    def is_default_languagecode(self, languagecode):
        """Is the provided languagecode the default language code?

        Note that this may be per domain etc. depending on the handler class.

        Args:
            languagecode (str): Language code.

        Returns:
            bool: True if the provided languagecode is the default.
        """
        return languagecode == self.default_languagecode

    @property
    def default_languagecode(self):
        """Get the default language code.

        Note that this may be per domain etc. depending on the handler class.

        Defaults to ``settings.LANGUAGE_CODE``.

        Returns:
            str: Default languagecode.
        """
        return active_i18n_url_translation.get_default_languagecode()

    @property
    def active_languagecode(self):
        """Get the active languagecode.

        Returns:
            str: The active languagecode.
        """
        return active_i18n_url_translation.get_active_languagecode()

    @property
    def active_languagecode_or_none_if_default(self):
        """Get the active languagecode, but returns None if the active languagecode is
        the default languagecode.

        Returns:
            str: The active languagecode, or None if the active languagecode is the default languagecode.
        """
        languagecode = self.active_languagecode
        if self.is_default_languagecode(languagecode):
            return None
        return languagecode

    @property
    def active_base_url(self):
        """Get the active base url.

        Returns:
            str: The active base url.
        """
        return active_i18n_url_translation.get_active_base_url()

    @classmethod
    def is_supported_languagecode(cls, languagecode):
        """Is the provided languagecode a supported languagecode?

        Args:
            languagecode (str): Language code.

        Returns:
            bool: True if the provided languagecode is supported.
        """
        return languagecode in cls.get_all_supported_languagecodes()

    def get_translated_label_for_languagecode(self, languagecode):
        """Get the translated label for the languagecode (the name of the language)
        in the currently active language.

        This defaults to the english name for the language fetched
        via ``django.utils.translation.get_language_info()``.

        This is typically used in subclasses that override :meth:`~.AbstractHandler.get_label_for_languagecode`
        and change to translated labels by default.

        Args:
            languagecode (str): Language code.

        Returns:
            str: Translated label for the languagecode.
        """
        language_info = get_language_info(languagecode)
        return language_info['name_translated']

    def get_untranslated_label_for_languagecode(self, languagecode):
        """Get the untranslated label for the languagecode (the name of the language).

        Should return a label for the languagecode in a commonly used
        language that most users of your site will understand.

        This defaults to the english name for the language fetched
        via ``django.utils.translation.get_language_info()``.

        This is the what the default implementation of :meth:`~.AbstractHandler.get_label_for_languagecode`
        uses.

        Args:
            languagecode (str): Language code.

        Returns:
            str: Unstranslated label for the languagecode.
        """
        language_info = get_language_info(languagecode)
        return language_info['name']

    def get_local_label_for_languagecode(self, languagecode):
        """Get the local label for the languagecode (the name of the language in that language).

        This defaults to the english name for the language fetched
        via ``django.utils.translation.get_language_info()``.

        This is typically used in subclasses that override :meth:`~.AbstractHandler.get_label_for_languagecode`
        and change to labels in the native language by default.

        Args:
            languagecode (str): Language code.

        Returns:
            str: Local label for the languagecode.
        """
        language_info = get_language_info(languagecode)
        return language_info['name_local']

    def get_label_for_languagecode(self, languagecode):
        """Get the label for the languagecode (the name of the language).

        Defaults to using :meth:`~.AbstractHandler.get_local_label_for_languagecode`.
        I.e.: We use the native/local translation of the language name as
        language label by default.

        Args:
            languagecode (str): Language code.

        Returns:
            str: Label for the languagecode.
        """
        return self.get_local_label_for_languagecode(languagecode)

    @classmethod
    def build_base_url_from_request(cls, request):
        return BaseUrl(request.build_absolute_uri('/'))

    @classmethod
    def _detect_activate_translation_kwargs_from_request(cls, request):
        active_base_url = cls.build_base_url_from_request(request=request)
        default_languagecode = cls.detect_default_languagecode(base_url=active_base_url)
        if cls.is_translatable_urlpath(active_base_url, request.path):
            current_languagecode = cls.detect_current_languagecode(base_url=active_base_url, request=request)
            if not current_languagecode or not cls.is_supported_languagecode(current_languagecode):
                current_languagecode = default_languagecode
        else:
            current_languagecode = default_languagecode
        active_language_urlpath_prefix = cls.get_urlpath_prefix_for_languagecode(
            base_url=active_base_url,
            languagecode=current_languagecode)
        return {
            'active_translation_languagecode': cls.get_translation_to_activate_for_languagecode(current_languagecode),
            'active_languagecode': current_languagecode,
            'default_languagecode': default_languagecode,
            'active_language_urlpath_prefix': active_language_urlpath_prefix,
            'active_base_url': active_base_url
        }

    @classmethod
    def activate_languagecode_from_request(cls, request):
        """Activate the detected languagecode.

        This is what :class:`ievv_opensource.ievv_i18n_url.middleware.LocaleMiddleware`
        uses to process the request.

        What this does:

        - Builds a base_url using ``request.build_absolute_uri('/')``.
        - Finds the default language code using :meth:`~.AbstractHandler.detect_default_languagecode`
        - Finds the current language code using :meth:`~.AbstractHandler.detect_current_languagecode`
        - Handles fallback to default languagecode if the current languagecode is unsupported or the requested
          url path is not not translatable (using :meth:`~.AbstractHandler.is_supported_languagecode`
          and :meth:`~.AbstractHandler.is_translatable_urlpath`).
        - Finds the language URL prefix :meth:`~.AbstractHandler.get_urlpath_prefix_for_languagecode`.
        - Activates the current language/translation using
          :func:`ievv_opensource.ievv_i18n_url.active_i18n_url_translation.activate`.

        .. warning::

            Do not override this method, and you should normally not call this method.
            I.e.: This is for the middleware.
        """
        active_i18n_url_translation.activate(**cls._detect_activate_translation_kwargs_from_request(request=request))

    ##################################################
    #
    # Methods designed to be overridden in subclasses:
    #
    ##################################################

    def get_icon_cssclass_for_languagecode(self, languagecode):
        """Get an icon CSS class for the language code.

        This is typically implemented on a per app basis. I.e.: The application
        creates a subclass of one of the built-in handlers and override this
        to provide icons for their supported languages. This is provided as
        part of the handler to make it possible to generalize things like
        rendering language selects with icons.

        This icon must be possible to use in HTML like this::

            <span class="ICON_CSS_CLASS_HERE"></span>

        Args:
            languagecode (str): Language code.

        Returns:
            str: Icon css class
        """
        return None

    def get_icon_svg_image_url_for_languagecode(self, languagecode):
        """Get an icon SVG image URL for the language code.

        This is typically implemented on a per app basis. I.e.: The application
        creates a subclass of one of the built-in handlers and override this
        to provide icons for their supported languages. This is provided as
        part of the handler to make it possible to generalize things like
        rendering language selects with icons.

        Args:
            languagecode (str): Language code.

        Returns:
            str: SVG image URL.
        """
        return None

    def build_absolute_url(self, path, languagecode=None, base_url=None):
        """Build absolute uri for the provided path within the provided languagecode.

        MUST be implemented in subclasses.

        .. note::

            Session based handlers will ignore the languagecode argument and just return
            the URL for the default translation. This is because all their translations live at the same URL.
            See the *A warning about session based translations* in the docs for more details.

        Args:
            path (str): The path (same format as HttpRequest.get_full_path()
                returns - e.g: ``"/my/path?option1&option2"``)
            languagecode (str, optional): The languagecode to build the URI for. Defaults to None, which
                means we build the URI within the current languagecode.
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL to reverse the path+languagecode in - see
                :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
                If this is ``None``, we use :obj:`~.AbstractHandler.active_base_url`.
        """
        raise NotImplementedError()

    def build_urlpath(self, path, languagecode=None, base_url=None):
        """Build URL path for the provided path within the provided languagecode.

        This is a compatibility layer to make it possible to work with older code
        that considers a URL path as fully qualified. Most handlers will just
        do nothing with the path, or prepend a prefix, but some handlers (those that work with multiple domains),
        will use the ``ievv_i18n_url_redirect_to_languagecode`` redirect view here to
        return an URL that will redirect the user to the correct URL.

        MUST be implemented in subclasses.

        .. note::

            Session based handlers will ignore the languagecode argument and just return
            the PATH for the default translation. This is because all their translations live at the same URL.
            See the *A warning about session based translations* in the docs for more details.

        Args:
            path (str): The path (same format as HttpRequest.get_full_path()
                returns - e.g: ``"/my/path?option1&option2"``)
            languagecode (str, optional): The languagecode to build the path for. Defaults to None, which
                means we build the URI within the current languagecode.
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL to reverse the path+languagecode in - see
                :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
                If this is ``None``, we use :obj:`~.AbstractHandler.active_base_url`.
        """
        raise NotImplementedError()

    @classmethod
    def transform_urlpath_to_languagecode(cls, base_url, path, from_languagecode, to_languagecode):
        """
        Transform the provided ``path`` (url path) into the provided languagecode.

        This is a utility method that is normally used by :meth:`~.AbstractHandler.transform_url_to_languagecode`
        to handle the ``translate_path=True`` argument. It SHOULD NOT be used directly.

        Raises:
            UrlTransformError: If the URL path can not be translated.
        Args:
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL of the provided ``path``. The result may be ment to
                live in a different ``base_url``, but that is not the responsibility of this
                method - that is the responsibility of :meth:`~.AbstractHandler.transform_url_to_languagecode`.
            path (str): The URL path to translate.
            from_languagecode (str): The languagecode to translate from.
            to_languagecode (str): The languagecode to translate to.
        Returns:
            str: The translated URL path.
        """
        current_language = translation.get_language()
        try:
            translation.activate(from_languagecode)
            url_parts = resolve(path)
            new_path = path
            translation.activate(to_languagecode)
            new_path = reverse(url_parts.view_name, args=url_parts.args, kwargs=url_parts.kwargs)
        except Resolver404 as resolver_404_error:
            raise UrlTransformError(str(resolver_404_error))
        except NoReverseMatch as no_reverse_match_error:
            raise UrlTransformError(str(no_reverse_match_error))
        finally:
            translation.activate(current_language)
        return str(new_path)

    @classmethod
    def transform_url_to_languagecode(cls, url, languagecode, translate_path=True):
        """Transform the provided url into the "same" url, but in the provided languagecode.

        MUST be implemented in subclasses.

        .. note::

            This is not possible to implement in a safe
            manner for session based handlers (I.e.: multiple translation live at the same URI),
            so for these kind of handler this method will just return the provided `url`.
            See the *A warning about session based translations* in the docs for more details.

        Args:
            url (str): The URL to transform.
            languagecode (str): The languagecode to transform the URL into.
            translate_path (bool): Translate the path? This means that the handler tries
                to convert the path to the translated path. E.g.: The path may be ``/about/contact-us``
                in english, but ``/om/kontakt-oss`` in norwegian, and with this argument set to ``True``,
                the handler will try its best to make that happen.
        """
        raise NotImplementedError()

    @classmethod
    def get_translation_to_activate_for_languagecode(cls, languagecode):
        """Get the languagecode to actually activate for the provided languagecode.

        Used by the middleware provided by `ievv_i18n_url` to activate the translation for the
        provided languagecode.

        This is here to make it possible for applications to have languages that has their
        own domains or URLs, but show translations from another language code. E.g.:
        you may have content in a language, but be OK with translation strings from another language.

        Returns:
            str: The languagecode to activate with the django translation system for the provided ``languagecode``.
            Defaults to the provided ``languagecode``.
        """
        return languagecode

    @classmethod
    def get_all_supported_languagecodes(cls):
        """Get ALL the supported language codes.

        I.e.: All supported languagecodes, not filtered by domain or anything.

        Defaults to the language codes in ``settings.LANGUAGES``.

        Returns:
            set: A set of the supported language codes.
        """
        return {l[0] for l in settings.LANGUAGES}

    @classmethod
    def get_supported_languagecodes(cls, base_url):
        """Get the supported language codes within the provided ``base_url``.

        Defaults to :meth:`.get_all_supported_languagecodes`, but will typically be
        overridden in handlers that deal with binding languagecodes to domains.

        Args:
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL - see :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
        Returns:
            set: A set of the supported language codes within the ``base_url``.
        """
        return cls.get_all_supported_languagecodes()

    @classmethod
    def has_multiple_supported_languages(cls, base_url):
        """Does the provided ``base_url`` support multiple languages?

        Defaults to checking if the length of :meth:`.get_supported_languagecodes` is more than 1.

        Args:
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL - see :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
        Returns:
            bool: True if we support multiple languages within the provided ``base_url``.
        """
        return len(cls.get_supported_languagecodes(base_url)) > 1

    def get_supported_languagecodes_in_active_base_url(self):
        """Get the supported language codes within the :obj:`~.AbstractHandler.active_base_url`.

        Defaults to calling :meth:`~.AbstractHandler.get_supported_languagecodes`, and you will
        normally NOT want to override this (override :meth:`~.AbstractHandler.get_supported_languagecodes` instead).

        Returns:
            set: A set of the supported language codes within the active base url.
        """
        return self.__class__.get_supported_languagecodes(base_url=self.active_base_url)

    def has_multiple_supported_languages_in_active_base_url(self):
        """Does the :obj:`~.AbstractHandler.active_base_url` support multiple languages?

        Defaults to calling :meth:`~.AbstractHandler.has_multiple_supported_languages`, and you will
        normally NOT want to override this (override :meth:`~.AbstractHandler.has_multiple_supported_languages` instead).

        Returns:
            bool: True if we support multiple languages within the active base url.
        """
        return self.__class__.has_multiple_supported_languages(base_url=self.active_base_url)

    # @classmethod
    # def strip_languagecode_from_urlpath(cls, base_url, languagecode, path):
    #     """Strip the languagecode from the URL path.

    #     Must be overridden in subclasses.

    #     Args:
    #         base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
    #             The base URL - see :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
    #         languagecode (str): The languagecode to strip out of ``path``.
    #         path (str): An URL path (e.g: ``/my/path``, ``/my/path?a=1&b=2``, etc.)
    #     Returns:
    #         str: The path without any languagecode prefix.
    #     """
    #     raise NotImplementedError()

    @classmethod
    def is_translatable_urlpath(self, base_url, path):
        """Is the provided URL path translatable within the current base_url?

        We default to consider the paths in settings.MEDIA_URL and settings.STATIC_URL as untranslatable.

        If this returns ``False``, the middleware will use the default translation when serving
        the path.

        If you subclass this, you should not write code that parses the querystring part of the
        path. This will not work as intended (will not work the same everywhere) since the middleware
        does not call this with the querystring included, but other code using this may pass in the path
        with the querystring.

        Args:
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL - see :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
            path (str): An URL path (e.g: ``/my/path``, ``/my/path?a=1&b=2``, etc.)
        Returns:
            bool: Is the provided URL translatable?
        """
        if settings.STATIC_URL and path.startswith(settings.STATIC_URL):
            return False
        if settings.MEDIA_URL and path.startswith(settings.MEDIA_URL):
            return False
        return True

    @classmethod
    def detect_preferred_languagecode_for_user(cls, user):
        """Detect the preferred languagecode for the provided user.

        This is normally NOT used by :meth:`~.AbstractHandler.detect_current_languagecode`
        except for handlers like the session handler where the URL does not change based
        on language code. E.g.: It would be strange to serve a language based on user
        preferences when the URL explicitly says what language code we are serving.

        This is mostly a convenience for management scripts, and a utility if you want to
        redirect users based on the preferred language code.

        Args:
            user: A user model object.

        Returns:
            str: The preferred language code for the provided user, or None.
            None means that the user has no preferred language code,
            or that the handler does not respect user preferences. Returns ``None`` by default.
        """
        return None

    @classmethod
    def get_languagecode_from_url(cls, url):
        """Get the languagecode from the provided ``url``.

        MUST be overridden in subclasses.

        Args:
            url (str): The URL to find the languagecode from.
        """
        raise NotImplementedError()

    @classmethod
    def detect_current_languagecode(cls, base_url, request):
        """Detect the current languagecode from the provided request and/or base_url.

        Used by the middleware provided by `ievv_i18n_url` find the current language code
        and set it on the current request.

        DO NOT USE THIS - it is for the middleware. Use :obj:`~.AbstractHandler.current_languagecode`
        or :meth:`~.AbstractHandler.get_languagecode_from_url`.

        You normally do not need to override this method in subclasses - it defaults to
        using :meth:`~.AbstractHandler.get_languagecode_from_url` with the full URL
        in the ``request`` as the url argument.

        If this returns None, it means that we should use the default languagecode. I.e.: Do you do not need to handle
        fallback to default languagecode when implementing this method in subclasses - just return None.

        Args:
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL - see :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
            request (django.http.HttpRequest): The HttpRequest

        Returns:
            str: The current languagecode or None (None means default languagecode is detected).
        """
        return cls.get_languagecode_from_url(request.build_absolute_uri())

    @classmethod
    def detect_default_languagecode(cls, base_url):
        """Detect the default languagecode for the provided base_url.

        This is here so that handlers can override it to support different default languagecode
        per domain or perhaps more fancy stuff based on the provided url.

        Args:
            base_url (django.http.HttpRequest):
                The base URL - e.g: https://example.com, http://www.example.com:8080, ...
                (NOT https://example.com/my/path, https://example.com/en, ...).

        Returns:
            str: The default languagecode. Defaults to ``settings.LANGUAGE_CODE``.
        """
        return settings.LANGUAGE_CODE

    @classmethod
    def get_urlpath_prefix_for_languagecode(cls, base_url, languagecode):
        """
        Get the URL path prefix for the provided languagecode within the provided base_url.

        Args:
            base_url (ievv_opensource.ievv_i18n_url.base_url.BaseUrl):
                The base URL - see :class:`ievv_opensource.ievv_i18n_url.base_url.BaseUrl` for more info.
            languagecode (str): The language code to find the prefix for.

        Returns:
            str: The url path prefix. **Can not** start or end with ``/``.
        """
        return ''



from django.urls import reverse

from .get_handler import get_handler


def i18n_reverse(viewname, urlconf=None, args=None, kwargs=None, current_app=None, languagecode=None):
    """
    Serves kind of the same use case as the ``django.urls.reverse`` function, but with i18n URL support, AND
    this function returns an absolute URL.

    The reason why it returns absolute URL is because i18n URLs may be based on domains, not just URL paths.

    NOTE: Session based `ievv_i18n_url` handlers will ignore the languagecode argument and just return
    the URL for the default translation. This is because all their translations live at the same URL.
    See the *A warning about session based translations* in the docs for more details.

    Args:
        viewname: See the docs for ``django.urls.reverse``.
        urlconf: See the docs for ``django.urls.reverse``. Defaults to None.
        args: See the docs for ``django.urls.reverse``. Defaults to None.
        kwargs: See the docs for ``django.urls.reverse``. Defaults to None.
        current_app: See the docs for ``django.urls.reverse``. Defaults to None.
        languagecode (str, optional): The languagecode to reverse the URL in. Defaults to None, which means
            we reverse the URL in the current languagecode.

    Returns:
        str: An URL.
    """
    path = reverse(viewname=viewname, urlconf=urlconf, args=args, kwargs=kwargs, current_app=current_app)
    return get_handler().build_absolute_url(path=path, languagecode=languagecode)


def transform_url_to_languagecode(url, to_languagecode, from_languagecode=None):
    """Takes an URL, and finds the URL to the same content within a different languagecode.

    NOTE: Session based `ievv_i18n_url` handlers will ignore the languagecode argument and just return
    provided url. This is because all their translations live at the same URL.
    See the *A warning about session based translations* in the docs for more details.

    Args:
        url (str): The URL to transform.
        to_languagecode (str): The languagecode to transform the url into.
        from_languagecode (str): The languagecode to transform the url from.

    Returns:
        str: The transformed URL. If from_languagecode and to_languagecode is the same,
            the provided ``url`` is returned unchanged.
    """
    return get_handler().transform_url_to_languagecode(url=url, to_languagecode=to_languagecode)



from django import test
from django.core.exceptions import ValidationError

from ievv_opensource.utils.validation_error_util import ValidationErrorUtil


class TestValidationErrorUtil(test.TestCase):
    def test_is_dict_based(self):
        self.assertFalse(ValidationErrorUtil(ValidationError('Error')).is_dict_based)
        self.assertFalse(ValidationErrorUtil(ValidationError(['Error1', 'Error2'])).is_dict_based)
        self.assertTrue(ValidationErrorUtil(ValidationError({'myfield': 'MyError'})).is_dict_based)

    def test_as_list(self):
        simpleerror = ValidationError('Error')
        self.assertEqual(len(ValidationErrorUtil(simpleerror).as_list()), 1)
        self.assertEqual(ValidationErrorUtil(simpleerror).as_list()[0].message, 'Error')

        listerror = ValidationError(['Error1', 'Error2'])
        self.assertEqual(len(ValidationErrorUtil(listerror).as_list()), 2)
        self.assertEqual(ValidationErrorUtil(listerror).as_list()[0].message, 'Error1')
        self.assertEqual(ValidationErrorUtil(listerror).as_list()[1].message, 'Error2')

        dicterror = ValidationError({'field1': ['Error1', 'Error2'], 'field2': 'Error3'})
        self.assertEqual(len(ValidationErrorUtil(dicterror).as_list()), 3)
        self.assertEqual(
            {validation_error.message for validation_error in ValidationErrorUtil(dicterror).as_list()},
            {'Error1', 'Error2', 'Error3'})

    def test_as_serializable_list(self):
        simpleerror = ValidationError('Error')
        self.assertEqual(len(ValidationErrorUtil(simpleerror).as_serializable_list()), 1)
        self.assertEqual(ValidationErrorUtil(simpleerror).as_serializable_list()[0], 'Error')

        listerror = ValidationError(['Error1', 'Error2'])
        self.assertEqual(len(ValidationErrorUtil(listerror).as_serializable_list()), 2)
        self.assertEqual(ValidationErrorUtil(listerror).as_serializable_list()[0], 'Error1')
        self.assertEqual(ValidationErrorUtil(listerror).as_serializable_list()[1], 'Error2')

        dicterror = ValidationError({'field1': ['Error1', 'Error2'], 'field2': 'Error3'})
        self.assertEqual(len(ValidationErrorUtil(dicterror).as_serializable_list()), 3)
        self.assertEqual(
            {validation_error for validation_error in ValidationErrorUtil(dicterror).as_serializable_list()},
            {'Error1', 'Error2', 'Error3'})

    def test_as_serializable_dict_simple(self):
        self.assertEqual(
            ValidationErrorUtil(ValidationError('Error')).as_serializable_dict(),
            {
                '__all__': ['Error']
            }
        )

    def test_as_serializable_dict_list(self):
        self.assertEqual(
            ValidationErrorUtil(ValidationError(['Error1', 'Error2'])).as_serializable_dict(),
            {
                '__all__': ['Error1', 'Error2']
            }
        )

    def test_as_serializable_dict_dict(self):
        self.assertEqual(
            ValidationErrorUtil(
                ValidationError({'field1': ['Error1', 'Error2'], 'field2': 'Error3'})
            ).as_serializable_dict(),
            {
                'field1': ['Error1', 'Error2'],
                'field2': ['Error3']
            }
        )

    def test_as_drf_validation_error_simple(self):
        drf_validation_error = ValidationErrorUtil(ValidationError('Error')).as_drf_validation_error()
        self.assertEqual(len(drf_validation_error.detail), 1)
        self.assertEqual(drf_validation_error.detail[0], 'Error')

    def test_as_drf_validation_error_simple_code(self):
        drf_validation_error = ValidationErrorUtil(ValidationError('Error', code='myerror')).as_drf_validation_error()
        self.assertEqual(drf_validation_error.detail[0].code, 'myerror')

    def test_as_drf_validation_error_list(self):
        drf_validation_error = ValidationErrorUtil(ValidationError(['Error1', 'Error2'])).as_drf_validation_error()
        self.assertEqual(len(drf_validation_error.detail), 2)
        self.assertEqual(drf_validation_error.detail[0], 'Error1')
        self.assertEqual(drf_validation_error.detail[1], 'Error2')

    def test_as_drf_validation_error_dict(self):
        drf_validation_error = ValidationErrorUtil(
            ValidationError({'field1': ['Error1', 'Error2'], 'field2': 'Error3'})
        ).as_drf_validation_error()
        self.assertEqual(set(drf_validation_error.detail.keys()), {'field1', 'field2'})
        self.assertEqual(drf_validation_error.detail['field1'][0], 'Error1')
        self.assertEqual(drf_validation_error.detail['field1'][1], 'Error2')
        self.assertEqual(drf_validation_error.detail['field2'][0], 'Error3')

    def test_as_drf_validation_error_dict_code(self):
        drf_validation_error = ValidationErrorUtil(
            ValidationError({
                'field1': [
                    ValidationError('Error1', code='customcode'),
                    'Error2'
                ],
                'field2': ValidationError('Error3', code='customcode2')
            })
        ).as_drf_validation_error()
        self.assertEqual(drf_validation_error.detail['field1'][0].code, 'customcode')
        self.assertEqual(drf_validation_error.detail['field1'][1].code, 'invalid')
        self.assertEqual(drf_validation_error.detail['field2'][0].code, 'customcode2')



import traceback
from django.contrib.postgres.fields import DateTimeRangeField

from django.core.exceptions import ValidationError
from rest_framework import serializers
from rest_framework.serializers import raise_errors_on_nested_writes
from rest_framework.utils import model_meta

from ievv_opensource.ievv_restframework_helpers.serializer_fields import DateTimeRangeSerializerField
from ievv_opensource.utils.validation_error_util import ValidationErrorUtil


class FullCleanModelSerializer(serializers.ModelSerializer):
    """
    A ``rest_framework.serializers.ModelSerializer`` subclass
    that calls full_clean() on the model instance before saving.
    """
    def clean_model_instance(self, instance):
        """
        Clean the model instance (full_clean()) and ensures any
        ``django.core.exceptions.ValidationError`` raised is re-raised as a
        ``rest_framework.exceptions.ValidationError``.
        """
        try:
            instance.full_clean()
        except ValidationError as e:
            raise ValidationErrorUtil(e).as_drf_validation_error()

    def create(self, validated_data):
        """
        We have a bit of extra checking around this in order to provide
        descriptive messages when something goes wrong, but this method is
        essentially just:

            return ExampleModel.objects.create(**validated_data)

        If there are many to many fields present on the instance then they
        cannot be set until the model is instantiated, in which case the
        implementation is like so:

            example_relationship = validated_data.pop('example_relationship')
            instance = ExampleModel.objects.create(**validated_data)
            instance.example_relationship = example_relationship
            return instance

        The default implementation also does not handle nested relationships.
        If you want to support writable nested relationships you'll need
        to write an explicit `.create()` method.
        """
        raise_errors_on_nested_writes('create', self, validated_data)

        ModelClass = self.Meta.model

        # Remove many-to-many relationships from validated_data.
        # They are not valid arguments to the default `.create()` method,
        # as they require that the instance has already been saved.
        info = model_meta.get_field_info(ModelClass)
        many_to_many = {}
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = validated_data.pop(field_name)

        instance = ModelClass(**validated_data)
        self.clean_model_instance(instance)
        try:
            instance.save()
        except TypeError:
            tb = traceback.format_exc()
            msg = (
                'Got a `TypeError` when calling `%s.save()`. '
                'This may be because you have a writable field on the '
                'serializer class that is not a valid argument to '
                '`%s()`. You may need to make the field '
                'read-only, or override the %s.create() method to handle '
                'this correctly.\nOriginal exception was:\n %s' %
                (
                    ModelClass.__name__,
                    ModelClass.__name__,
                    self.__class__.__name__,
                    tb
                )
            )
            raise TypeError(msg)

        # Save many-to-many relationships after the instance is created.
        if many_to_many:
            for field_name, value in many_to_many.items():
                field = getattr(instance, field_name)
                field.set(value)

        return instance

    def update(self, instance, validated_data):
        raise_errors_on_nested_writes('update', self, validated_data)
        info = model_meta.get_field_info(instance)

        # Simply set each attribute on the instance, and then save it.
        # Note that unlike `.create()` we don't need to treat many-to-many
        # relationships as being a special case. During updates we already
        # have an instance pk for the relationships to be associated with.
        for attr, value in validated_data.items():
            if attr in info.relations and info.relations[attr].to_many:
                field = getattr(instance, attr)
                field.set(value)
            else:
                setattr(instance, attr, value)

        self.clean_model_instance(instance)
        instance.save()
        return instance


class IevvModelSerializer(FullCleanModelSerializer):
    def build_field(self, field_name, info, model_class, nested_depth):
        result = super().build_field(field_name, info, model_class, nested_depth)
        if isinstance(model_class._meta.get_field(field_name), DateTimeRangeField):
            return DateTimeRangeSerializerField, result[1]
        return result



import json
from django.utils.html import format_html
from django.contrib import admin
from .models import IevvLoggingEventBase, IevvLoggingEventItem


@admin.register(IevvLoggingEventBase)
class IevvLoggingEventBaseAdmin(admin.ModelAdmin):
    list_display = [
        'slug',
        'last_started',
        'success_run',
        'time_spent_in_seconds',
    ]
    search_fields = [
        'slug'
    ]
    list_filter = [
        'last_started',
        'success_run',
    ]
    readonly_fields = [
        'slug',
        'last_started',
        'last_finished',
        'time_spent_in_seconds',
        'time_spent',
        'success_run',
    ]


@admin.register(IevvLoggingEventItem)
class IevvLoggingEventItemAdmin(admin.ModelAdmin):
    list_display = [
        'logging_base',
        'success_run',
        'created_datetime',
        'time_spent_in_seconds',
    ]
    search_fields = [
        'logging_base'
    ]
    list_filter = [
        'logging_base',
        'success_run',
        'created_datetime',
    ]
    readonly_fields = [
        'logging_base',
        'get_data_pretty',
        'created_datetime',
        'start_datetime',
        'end_datetime',
        'time_spent',
        'time_spent_in_seconds',
        'success_run',
    ]
    exclude = [
        'data',
    ]

    def get_data_pretty(self, obj):
        return format_html('<pre>{}</pre>', json.dumps(obj.data, indent=2, sort_keys=True))
    get_data_pretty.short_description = 'Data pretty formatted'



import os

from django.conf import settings
from django.core import management
from django.core.management.base import BaseCommand


def get_dumpdata_filepath():
    return os.path.join(settings.IEVVTASKS_DUMPDATA_DIRECTORY,
                        'default.sql')


class Command(BaseCommand):
    help = 'Dump the development database as SQL.'

    def handle(self, *args, **options):
        outfile = get_dumpdata_filepath()
        if not os.path.exists(settings.IEVVTASKS_DUMPDATA_DIRECTORY):
            os.makedirs(settings.IEVVTASKS_DUMPDATA_DIRECTORY)
        management.call_command('dbdev_createdump', outfile)
        self.stdout.write('Dumped the database into {}.'.format(
            outfile))



from django.conf import settings
from django.contrib.staticfiles.templatetags import staticfiles


def get_setting():
    return getattr(settings, 'IEVV_STATICFILES_AUTOGZIP', None) or {}


def is_enabled_for_context(autogzip_context):
    return get_setting().get(autogzip_context, False)


def static_path(path, autogzip_context):
    """
    If the ``autogzip_context`` is set to True in the 
    django setting IEVV_STATICFILES_AUTOGZIP (a dict), the provided
    path is suffixed with ``.gz`` and returned. Otherwise,
    ``path`` is returned unmodified.
    """
    if is_enabled_for_context(autogzip_context):
        return f'{path}.gz'
    return path


def static(path, autogzip_context):
    """
    Just like ``django.contrib.staticfiles.templatetags.static()``,
    but if the ``autogzip_context`` is set to True in the 
    django setting IEVV_STATICFILES_AUTOGZIP (a dict), the provided
    path is suffixed with ``.gz``.
    """
    return staticfiles.static(static_path(path, autogzip_context))



import os
import sys
import subprocess


def open_file_with_default_os_opener(filename):
    if sys.platform == "win32":
        os.startfile(filename)
    else:
        if sys.platform == "darwin":
            opener = "open"
        else:
            opener = "xdg-open"
        subprocess.call([opener, filename])



import gzip


class GzipCompressMixin:
    def gzip_compression_enabled(self):
        return False

    def gzip_gzip_compresslevel(self):
        return 9

    def get_filepaths_to_gzip(self):
        return []

    def make_gzip_outfilepath(self, sourcepath):
        return f'{sourcepath}.gz'

    def gzip_compress_file(self, sourcepath):
        gzip_outfilepath = self.make_gzip_outfilepath(sourcepath)
        sourcecode = open(sourcepath, 'rb').read()
        gzipped_data = gzip.compress(sourcecode, compresslevel=self.gzip_gzip_compresslevel())
        open(gzip_outfilepath, 'wb').write(gzipped_data)
        self.add_deferred_success(f'GZIP compressed {sourcepath} -> {gzip_outfilepath}.')

    def gzip_compress_files(self):
        if not self.gzip_compression_enabled():
            return
        for filepath in self.get_filepaths_to_gzip():
            self.gzip_compress_file(sourcepath=filepath)



from django import test

from ievv_opensource.ievv_sms.backends import pswin


@test.override_settings(PSWIN_DEFAULT_COUNTRY_CODE='47')
class TestBackend(test.TestCase):
    def test_clean_phone_number__no_country_code(self):
        backend = pswin.Backend(
            phone_number='x',
            message='x')
        self.assertEqual(
            backend.clean_phone_number('12345678'),
            '4712345678')

    def test_clean_phone_number__plus_country_code(self):
        backend = pswin.Backend(
            phone_number='x',
            message='x')
        self.assertEqual(
            backend.clean_phone_number('+9912345678'),
            '9912345678')

    def test_clean_phone_number__00_country_code(self):
        backend = pswin.Backend(
            phone_number='x',
            message='x')
        self.assertEqual(
            backend.clean_phone_number('009912345678'),
            '9912345678')

    def test_clean_phone_number__whitespace(self):
        backend = pswin.Backend(
            phone_number='x',
            message='x')
        self.assertEqual(
            backend.clean_phone_number(' 123 45 678 '),
            '4712345678')

    @test.override_settings(PSWIN_SENDER='test')
    def test_pswin_sender_from_setting(self):
        backend = pswin.Backend(
            phone_number='12345678',
            message='test')
        self.assertEqual(backend.pswin_sender, 'test')

    def test_pswin_sender_from_send_as(self):
        backend = pswin.Backend(
            phone_number='12345678',
            message='test',
            send_as='A')
        self.assertEqual(backend.pswin_sender, 'A')

    def test_pswin_sender_from_pswin_sender_arg(self):
        backend = pswin.Backend(
            phone_number='12345678',
            message='test',
            pswin_sender='A')
        self.assertEqual(backend.pswin_sender, 'A')

    def test_pswin_sender_pswin_sender_arg_before_send_as(self):
        backend = pswin.Backend(
            phone_number='12345678',
            message='test',
            send_as='A',
            pswin_sender='B')
        self.assertEqual(backend.pswin_sender, 'B')

    @test.override_settings(PSWIN_SENDER='A')
    def test_pswin_sender_send_as_before_setting(self):
        backend = pswin.Backend(
            phone_number='12345678',
            message='test',
            send_as='B')
        self.assertEqual(backend.pswin_sender, 'B')



import urllib

from django import test
from ievv_opensource.ievv_i18n_url.base_url import BaseUrl


@test.override_settings(IEVV_I18N_URL_FALLBACK_BASE_URL='https://www.example.com')
class TestBaseUrl(test.TestCase):
    def test_parsed_url_none(self):
        self.assertEqual(BaseUrl().parsed_url.geturl(), 'https://www.example.com')

    def test_parsed_url_string(self):
        self.assertEqual(BaseUrl('https://string.example.com/x/y').parsed_url.geturl(), 'https://string.example.com')

    def test_parsed_url_parseresult(self):
        self.assertEqual(BaseUrl(urllib.parse.urlparse('https://parsed.example.com/x/y')).parsed_url.geturl(), 'https://parsed.example.com')

    def test_scheme(self):
        self.assertEqual(BaseUrl('https://example.com').scheme, 'https')
        self.assertEqual(BaseUrl('http://example.com').scheme, 'http')

    def test_netloc(self):
        self.assertEqual(BaseUrl('https://example.com').netloc, 'example.com')
        self.assertEqual(BaseUrl('https://example.com:8080').netloc, 'example.com:8080')

    def test_hostname(self):
        self.assertEqual(BaseUrl('https://example.com').hostname, 'example.com')
        self.assertEqual(BaseUrl('https://example.com:8080').hostname, 'example.com')

    def test_port(self):
        self.assertEqual(BaseUrl('https://example.com').port, None)
        self.assertEqual(BaseUrl('https://example.com:8080').port, 8080)

    def test_build_absolute_url(self):
        self.assertEqual(
            BaseUrl('https://example.com').build_absolute_url('/my/path?x=2'),
            'https://example.com/my/path?x=2')
        self.assertEqual(
            BaseUrl('https://example.com/a/b').build_absolute_url('/my/path?x=2'),
            'https://example.com/my/path?x=2')
        self.assertEqual(
            BaseUrl('https://example.com').build_absolute_url('https://other.example.com/a/b'),
            'https://other.example.com/a/b')



import re

from django import urls
from django.conf.urls import url
from django.conf import settings

from ievv_opensource.ievv_i18n_url import active_i18n_url_translation
from ievv_opensource.ievv_i18n_url.views import RedirectToLanguagecodeView

#
# Note: Basically copy and tuning from https://github.com/django/django/blob/1.11.29/django/conf/urls/i18n.py
#
# When updating for Django 2+, see https://github.com/django/django/blob/master/django/conf/urls/i18n.py
#


def i18n_patterns(*urls, include_redirect_view=True):
    """
    Adds the language code prefix to every URL pattern within this
    function. This may only be used in the root URLconf, not in an included
    URLconf.
    """
    if not settings.USE_I18N:
        return list(urls)
    return [
        url(r'^_ievv-i18n-redirect-to/(?P<languagecode>[a-zA-Z-]+)/(?P<path>.*)$',
            RedirectToLanguagecodeView.as_view(),
            name='ievv_i18n_url_redirect_to_languagecode'),
        I18nRegexURLResolver(list(urls)),
    ]


class I18nRegexURLResolver(urls.RegexURLResolver):
    """
    A URL resolver that always matches the active language code as URL prefix.

    Rather than taking a regex argument, we just override the ``regex``
    function to always return the active language-code as regex.
    """
    def __init__(
        self, urlconf_name, default_kwargs=None, app_name=None, namespace=None,
        prefix_default_language=False
    ):
        super().__init__(None, urlconf_name, default_kwargs, app_name, namespace)
        self.prefix_default_language = prefix_default_language

    @property
    def regex(self):
        language_code = active_i18n_url_translation.get_active_languagecode()

        if language_code not in self._regex_dict:
            urlpath_prefix = active_i18n_url_translation.get_active_language_urlpath_prefix()
            if urlpath_prefix:
                regex_string = '^%s/' % urlpath_prefix
            else:
                regex_string = ''
            self._regex_dict[language_code] = re.compile(regex_string, re.UNICODE)
        return self._regex_dict[language_code]



import traceback
from datetime import datetime, timedelta

from django import test
from ievv_opensource.ievv_logging.models import IevvLoggingEventBase, IevvLoggingEventItem
from ievv_opensource.ievv_logging.utils import IevvLogging, getDuration


class TestIevvLogging(test.TestCase):

    def test_begin(self):
        ievvlogging = IevvLogging('foo_bar')
        ievvlogging.begin()
        self.assertEqual(1, IevvLoggingEventBase.objects.count())
        self.assertEqual(1, IevvLoggingEventItem.objects.count())

    def test_finish(self):
        ievvlogging = IevvLogging('foo_bar')
        ievvlogging.begin()
        self.assertEqual(1, IevvLoggingEventBase.objects.count())
        self.assertEqual(1, IevvLoggingEventItem.objects.count())
        ievvlogging.finish()
        self.assertEqual(1, IevvLoggingEventBase.objects.count())
        self.assertEqual(1, IevvLoggingEventItem.objects.count())

    def test_that_no_duplicates_of_slug_is_created(self):
        ievvlogging = IevvLogging('foo_bar')
        ievvlogging.begin()
        ievvlogging2 = IevvLogging('foo_bar')
        ievvlogging2.begin()
        self.assertEqual(1, IevvLoggingEventBase.objects.count())

    def test_each_logging_gives_another_row_in_item_model(self):
        ievvlogging = IevvLogging('foo_bar')
        ievvlogging.begin()
        ievvlogging.finish()
        ievvlogging2 = IevvLogging('foo_bar')
        ievvlogging2.begin()
        ievvlogging2.finish()
        self.assertEqual(1, IevvLoggingEventBase.objects.count())
        self.assertEqual(2, IevvLoggingEventItem.objects.count())

    def test_jsondata_populating(self):
        ievvlogging = IevvLogging('foo_bar')
        ievvlogging.begin()
        a_dictionary = {'a': 'b'}
        ievvlogging.finish(
            foo=1,
            bar=2,
            **a_dictionary
        )
        self.assertTrue('a' in IevvLoggingEventItem.objects.first().data)
        self.assertEqual(3, len(IevvLoggingEventItem.objects.first().data))

    def test_success_run_is_NOT_set(self):
        ievvlogging = IevvLogging('foo')
        ievvlogging.begin()
        ievvlogging.finish()
        self.assertTrue(IevvLoggingEventItem.objects.first().success_run)
        self.assertIsNotNone(IevvLoggingEventItem.objects.first().success_run)

    def test_success_run_is_false(self):
        ievvlogging = IevvLogging('foo')
        ievvlogging.begin()
        ievvlogging.finish(error_occured=True)
        self.assertFalse(IevvLoggingEventItem.objects.first().success_run)

    def test_success_is_false_and_exception_msg_is_saved_in_data(self):
        ievvlogging = IevvLogging('foo')
        ievvlogging.begin()
        try:
            {'foo': 11}['bar']
        except Exception:
            ievvlogging.finish(
                error_occured=True,
                error=traceback.format_exc()
            )
        else:
            ievvlogging.finish()
        loggingitem = IevvLoggingEventItem.objects.first()
        self.assertFalse(loggingitem.success_run)
        self.assertTrue('error' in loggingitem.data)
        self.assertTrue('KeyError' in loggingitem.data['error'])

    def test_NO_error_occured_is_set_and_exception_msg_is_saved_in_data(self):
        # same test as above, but with no error
        ievvlogging = IevvLogging('foo')
        ievvlogging.begin()
        try:
            {'foo': 11}['foo']
        except Exception:
            ievvlogging.finish(
                error_occured=True,
                error=traceback.format_exc()
            )
        else:
            ievvlogging.finish()
        loggingitem = IevvLoggingEventItem.objects.first()
        self.assertTrue(loggingitem.success_run)
        self.assertFalse('error' in loggingitem.data)

    def test_duration_less_than_a_second(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(milliseconds=100)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        self.assertEqual('Less than a second', duration)

    def test_duration_some_seconds(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(seconds=33)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        duration_seconds = getDuration(from_dt=from_time, to_dt=to_time, interval='seconds')
        self.assertEqual('0 hours, 0 minutes and 33 seconds', duration)
        self.assertEqual(33, duration_seconds)

    def test_duration_some_minutes(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(minutes=45)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        duration_seconds = getDuration(from_dt=from_time, to_dt=to_time, interval='seconds')
        self.assertEqual('0 hours, 45 minutes and 0 seconds', duration)
        self.assertEqual(45*60, duration_seconds)

    def test_duration_some_minutes_and_seconds(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(minutes=45)
        from_time = from_time - timedelta(seconds=15)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        self.assertEqual('0 hours, 45 minutes and 15 seconds', duration)

    def test_duration_some_hours(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(hours=5)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        duration_seconds = getDuration(from_dt=from_time, to_dt=to_time, interval='seconds')
        self.assertEqual('5 hours, 0 minutes and 0 seconds', duration)
        self.assertEqual(5*60*60, duration_seconds)

    def test_duration_some_hours_minutes_and_seconds(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(hours=5)
        from_time = from_time - timedelta(minutes=22)
        from_time = from_time - timedelta(seconds=15)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        self.assertEqual('5 hours, 22 minutes and 15 seconds', duration)

    def test_duration_some_days_hours_minutes_and_seconds(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(days=15)
        from_time = from_time - timedelta(hours=22)
        from_time = from_time - timedelta(minutes=22)
        from_time = from_time - timedelta(seconds=15)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        self.assertEqual('15 days, 22 hours, 22 minutes and 15 seconds', duration)

    def test_duration_some_months_days_hours_minutes_and_seconds(self):
        to_time = datetime.now()
        from_time = to_time - timedelta(days=365*2)
        from_time = from_time - timedelta(hours=2)
        from_time = from_time - timedelta(minutes=2)
        from_time = from_time - timedelta(seconds=2)
        duration = getDuration(from_dt=from_time, to_dt=to_time)
        self.assertTrue('years' in duration)





from __future__ import print_function

from ievv_opensource.ievv_sms import sms_registry


class Backend(sms_registry.AbstractSmsBackend):
    """
    Backend that just prints the phone number and
    message. Does not send an SMS.

    Useful for development.

    The backend_id for this backend is ``debugprint``.
    """
    @classmethod
    def get_backend_id(cls):
        return 'debugprint'

    def send(self):
        print(
            '{border}\n'
            'DEBUG PRINT SMS:\n'
            'Send as: {send_as}\n'
            'Phone number: {phone_number}\n'
            'Message: {message}\n'
            'Message length: {message_length}\n'
            '{border}\n'.format(
                send_as=self.send_as,
                phone_number=self.cleaned_phone_number,
                message=self.cleaned_message,
                message_length=len(self.cleaned_message),
                border='-' * 70
            )
        )


class Latin1Backend(Backend):
    """
    Just like :class:`.Backend`, but this backend crashes
    if you try to send any characters outside the latin1
    charset.

    The backend_id for this backend is ``debugprint_latin1``.

    Very useful when the targeted production backend
    only supports latin1.
    """
    @classmethod
    def get_backend_id(cls):
        return 'debugprint_latin1'

    def clean_message(self, message):
        return message.encode('latin1').decode('latin1')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.test import TestCase

from ievv_opensource.utils.text.ievv_slugify import ievv_slugify


class TestIevvSlugify(TestCase):
    def test_ievv_slugify_e(self):
        self.assertEqual('a', ievv_slugify("æ"))

    def test_ievv_slugify_o(self):
        self.assertEqual('o', ievv_slugify("ø"))

    def test_ievv_slugify_a(self):
        self.assertEqual('a', ievv_slugify("å"))

    def test_ievv_slugify_word(self):
        self.assertEqual('halla-assen-gare-a-do', ievv_slugify('Hællæ Åssen gåre æ dø!'))

    def test_ievv_slugify_word_with_settings_map(self):
        character_replace_map = {
            'æ': 'ae',
            'ø': 'oe',
            'å': 'aa'
        }
        with self.settings(IEVV_SLUGIFY_CHARACTER_REPLACE_MAP=character_replace_map):
            self.assertEqual('haellae-assen-gaare-ae-doe', ievv_slugify('Hællæ Åssen gåre æ dø!'))



try:
    from email import message_from_bytes
except ImportError:
    # For python 2.x compatibility
    from future.backports.email import message_from_bytes

from collections import OrderedDict

from django.db import models


class DevelopEmail(models.Model):
    created_datetime = models.DateTimeField(auto_now_add=True)
    subject = models.TextField(null=False, blank=True, default='')
    from_email = models.TextField(
        null=False, blank=True, default='',
        db_index=True)
    to_emails = models.TextField(null=False, blank=True, default='')
    raw_message = models.TextField(null=False, blank=True, default='')

    class Meta:
        ordering = ['-created_datetime']

    @property
    def subject_with_fallback(self):
        return self.subject.strip() or '(No subject)'

    @property
    def message(self):
        if not hasattr(self, '_message'):
            self._message = message_from_bytes(self.raw_message.encode('utf-8'))
        return self._message

    def _get_message(self, content_type, allow_not_multipart_fallback=False):
        if self.message.is_multipart():
            for part in self.message.walk():
                content_charset = part.get_content_charset() or 'utf-8'

                part_content_type = part.get_content_type()
                part_content_disposition = str(part.get('Content-Disposition'))

                if part_content_type == content_type and 'attachment' not in part_content_disposition:
                    return part.get_payload(decode=True).decode(content_charset)
            return None
        elif allow_not_multipart_fallback:
            content_charset = self.message.get_content_charset() or 'utf-8'
            return self.message.get_payload(decode=True).decode(content_charset)
        return None

    @property
    def plaintext_message(self):
        return self._get_message(content_type='text/plain', allow_not_multipart_fallback=True)

    @property
    def html_message(self):
        return self._get_message(content_type='text/html', allow_not_multipart_fallback=False)

    def iter_attachment_messages(self):
        if self.message.is_multipart():
            for part in self.message.walk():
                part_content_disposition = str(part.get('Content-Disposition'))

                if 'attachment' in part_content_disposition:
                    yield part

    def iter_attachment_data(self, include_payload=True):
        for attachment_message in self.iter_attachment_messages():
            data = {
                'filename': attachment_message.get_filename(),
                'content_type': attachment_message.get_content_type(),
                'payload': attachment_message.get_payload(decode=True)
            }
            if include_payload:
                data['payload'] = attachment_message.get_payload(decode=True)
            yield data

    def get_attachment_dict(self, include_payload=False):
        attachmentsdict = OrderedDict()
        for attachmentdata in self.iter_attachment_data(include_payload=include_payload):
            if attachmentdata['filename']:
                attachmentsdict[attachmentdata['filename']] = attachmentdata
        return attachmentsdict

    def __str__(self):
        return '#{} - {}'.format(self.id, self.subject_with_fallback)



import sys
import time

from django.conf import settings

from ievv_opensource.utils.ievvdevrun.runnables import base


class RunnableThread(base.ShellCommandRunnableThread):
    """
    Django-DBdev run database runnable thread.

    Examples:

        You can just add it to your Django development settings with::

            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': ievvdevrun.config.RunnableThreadList(
                    ievvdevrun.runnables.dbdev_runserver.RunnableThread()
                )
            }

    """

    def get_logger_name(self):
        return 'Django-dbdev database server: {!r}'.format(
            settings.DATABASES['default']
        )

    def get_command_config(self):
        return {
            'executable': sys.executable,
            'args': ['manage.py', 'dbdev_startserver']
        }

    def dbdev_stopserver(self):
        self.run_shell_command(sys.executable,
                               args=['manage.py', 'dbdev_stopserver'])

    def start(self):
        super(RunnableThread, self).start()
        time.sleep(4)  # Block to give the server time to start.

    def run(self):
        self.dbdev_stopserver()
        super(RunnableThread, self).run()

    def stop(self):
        super(RunnableThread, self).stop()
        self.dbdev_stopserver()



import os

from django.apps import apps


class FilePathInterface(object):
    """
    Base interface for file path objects.

    We provide subclasses if this interface with different use cases:

    - :class:`.SourcePath`: Use this to specify a source file or folder
        in the sources directory of a :class:`ievv_opensource.utils.ievvbuildstatic.config.App`.
    - :class:`.DestinationPath`: Use this to specify a destination file or folder
        in the sources directory of a :class:`ievv_opensource.utils.ievvbuildstatic.config.App`.
    - :class:`.AbsoluteFilePath`: Use this to specify a file or folder that is
        not organized using the ievvbuildstatic directory layout.
    """
    @property
    def abspath(self):
        """
        Property that returns the absolute path to the file (or directory).

        Must be overridden in subclasses.
        """
        raise NotImplementedError()


class AbsoluteFilePath(FilePathInterface):
    """
    Absolute file path.

    Works just like :func:`os.path.join`, and we just
    assume that you provide a path that is absolute.
    """
    def __init__(self, *path):
        """

        Args:
            *path: One or more strings that make up an absolute path
                when joined with ``os.path.join(*path)``.

        Returns:

        """
        self._path = path

    @property
    def abspath(self):
        return os.path.join(*self._path)


class AbstractDjangoAppPath(FilePathInterface):
    """
    Abstract base class for file paths within a Django app.
    """
    def get_approot_path(self, appname):
        appconfig = apps.get_app_config(appname)
        return appconfig.path

    def get_approot_relative_path(self, appname, *path):
        return os.path.join(self.get_approot_path(appname=appname), *path)

    @property
    def abspath(self):
        raise NotImplementedError()


class SourcePath(AbstractDjangoAppPath):
    """
    A path to a file or directory within the source directory of
    a :class:`ievv_opensource.utils.ievvbuildstatic.config.App`.

    Assumes that the sourcefolder of the App is ``"staticsources"`` (the default).
    """
    def __init__(self, appname, *path):
        """
        Args:
            appname: The name of the Django app.
            *path: The path relative to the source directory of the
                :class:`ievv_opensource.utils.ievvbuildstatic.config.App`
                identified by ``appname``. Same format as :func:`os.path.join`.
        """
        self.appname = appname
        self.path = path

    @property
    def abspath(self):
        return self.get_approot_relative_path(self.appname, 'staticsources', self.appname, *self.path)


class DestinationPath(AbstractDjangoAppPath):
    """
    A path to a file or directory within the destination directory of
    a :class:`ievv_opensource.utils.ievvbuildstatic.config.App`.

    Assumes that the destinationfolder of the App is ``"static"`` (the default).
    """
    def __init__(self, appname, version, *path):
        """
        Args:
            appname: The name of the Django app.
            version: The version of the app.
            *path: The path relative to the destination directory of the
                :class:`ievv_opensource.utils.ievvbuildstatic.config.App`
                identified by ``appname``. Same format as :func:`os.path.join`.
        """
        self.appname = appname
        self.version = version
        self.path = path

    @property
    def abspath(self):
        return self.get_approot_relative_path(self.appname, 'static', self.appname, self.version, *self.path)


class AppPath(AbstractDjangoAppPath):
    """
    A path to a file or directory within the root directory of
    a :class:`ievv_opensource.utils.ievvbuildstatic.config.App`.
    """
    def __init__(self, appname, *path):
        """
        Args:
            appname: The name of the Django app.
            *path: The path relative to the root directory of the
                :class:`ievv_opensource.utils.ievvbuildstatic.config.App`
                identified by ``appname``. Same format as :func:`os.path.join`.
        """
        self.appname = appname
        self.path = path

    @property
    def abspath(self):
        return self.get_approot_relative_path(self.appname, *self.path)



import sys
import traceback

from django.core.management.base import BaseCommand

from ievv_opensource.ievv_logging.utils import IevvLogging
from ievv_opensource.ievv_logging.models import IevvLoggingEventBase


def delete_older_than_the_last_100():
    logmessages = []
    for loggingeventbase in IevvLoggingEventBase.objects.all():
        last_created_item_ids = loggingeventbase.ievvloggingeventitem_set.all()[:100].values_list('id', flat=True)
        items = loggingeventbase.ievvloggingeventitem_set.exclude(pk__in=list(last_created_item_ids))
        logmsg = f'Deleting {items.count()} for {loggingeventbase.slug}'
        sys.stdout.write(f'{logmsg}\n')
        logmessages.append(logmsg)
        items.delete()
    return logmessages


class Command(BaseCommand):
    help = 'Deletes IevvLoggingEventItem rows for each IevvLoggingEventBase except the last 100.'

    def handle(self, *args, **options):
        ievvlogging = IevvLogging(__name__)
        ievvlogging.begin()
        try:
            logmessages = delete_older_than_the_last_100()
        except Exception:
            ievvlogging.finish(
                error_occured=True,
                error=traceback.format_exc()
            )
        else:
            ievvlogging.finish(logmessages=logmessages)



import datetime
import decimal

import arrow
from django import test
from django.conf import settings
from model_mommy import mommy

from ievv_opensource.utils import ievv_json


class TestIevvJSON(test.TestCase):
    def test_datetime(self):
        data = {'somedatetime': arrow.get(datetime.datetime(2019, 12, 23, 23, 50), 'Asia/Tokyo').datetime}
        encoded = ievv_json.dumps(data, sort_keys=True)
        self.assertEquals(encoded, '{"somedatetime": {"__type__": "datetime", "value": "2019-12-23 23:50:00+09:00"}}')
        decoded = ievv_json.loads(encoded)
        self.assertDictEqual(decoded, data)

    def test_date(self):
        data = {'somedate': datetime.date(2019, 12, 24)}
        encoded = ievv_json.dumps(data, sort_keys=True)
        self.assertEquals(encoded, '{"somedate": {"__type__": "date", "value": "2019-12-24"}}')
        decoded = ievv_json.loads(encoded)
        self.assertDictEqual(decoded, data)

    def test_time(self):
        data = {'sometime': datetime.time(12, 40, 38, 123)}
        encoded = ievv_json.dumps(data, sort_keys=True)
        self.assertEquals(encoded, '{"sometime": {"__type__": "time", "value": "12:40:38.000123"}}')
        decoded = ievv_json.loads(encoded)
        self.assertDictEqual(decoded, data)

    def test_decimal(self):
        data = {'somedecimal': decimal.Decimal('10.2')}
        encoded = ievv_json.dumps(data, sort_keys=True)
        self.assertEquals(encoded, '{"somedecimal": {"__type__": "decimal", "value": "10.2"}}')
        decoded = ievv_json.loads(encoded)
        self.assertDictEqual(decoded, data)

    def test_djangomodel(self):
        user = mommy.make(settings.AUTH_USER_MODEL)
        data = {'someuser': user}
        encoded = ievv_json.dumps(data, sort_keys=True)
        self.assertEquals(
            encoded,
            '{"someuser": {"__model__": "auth.User", "__type__": "djangomodel", "value": %s}}' % (user.id,))
        decoded = ievv_json.loads(encoded)
        self.assertEqual(decoded, {
            'someuser': {'__model__': 'auth.User', '__type__': 'djangomodel', 'value': user.id}})



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.utils.text import slugify
from django.conf import settings


def ievv_slugify(value):
    """
    Replaces norwegian characters before slugify

    Args:
        value:

    Returns:
        slugified value

    """
    character_replace_map = {
        'æ': 'a',
        'ø': 'o',
        'å': 'a'
    }
    if hasattr(settings, 'IEVV_SLUGIFY_CHARACTER_REPLACE_MAP'):
        character_replace_map = settings.IEVV_SLUGIFY_CHARACTER_REPLACE_MAP
    for character, replace_character in character_replace_map.items():
        value = value.replace(character, replace_character)
    return slugify(value)



from django.utils.functional import cached_property
import urllib.parse
from ievv_opensource.ievv_i18n_url.i18n_url_settings import get_fallback_base_url_setting


class BaseUrl:
    """
    Defines ievv_i18n_url base url.

    The base URL is the URL to the root of the domain e.g: https://example.com, http://www.example.com:8080, etc.
    (NOT https://example.com/my/path, https://example.com/en, etc.).

    The constructor can take any valid absolute URL, or None
    (in which case it falls back on the IEVV_I18N_URL_FALLBACK_BASE_URL setting),
    but all the methods and properties work on a ``urllib.parse.ParseResult`` that
    does not have any of the URL parts after
    """
    def __init__(self, url_or_urllib_parseresult=None):
        self.url_or_urllib_parseresult = url_or_urllib_parseresult or get_fallback_base_url_setting()

    @cached_property
    def parsed_url(self):
        """Get the parsed URL.

        The returned URL only has scheme+domain+port (e.g.: scheme+netloc).

        Returns:
            urllib.parse.ParseResult: A parsed URL on the same format as urllib.parse.urlparse returns.
        """
        if isinstance(self.url_or_urllib_parseresult, urllib.parse.ParseResult):
            parsed_url = self.url_or_urllib_parseresult
        else:
            parsed_url = urllib.parse.urlparse(self.url_or_urllib_parseresult)
        return urllib.parse.ParseResult(
            scheme=parsed_url.scheme,
            netloc=parsed_url.netloc,
            path='', params='', query='', fragment='')

    @property
    def scheme(self):
        """Shortcut for parsed_url.scheme.

        See :obj:`~.BaseUrl.parsed_url`.

        Returns:
            str: The url scheme (e.g.: https, http, ...)
        """
        return self.parsed_url.scheme

    @property
    def netloc(self):
        """Shortcut for parsed_url.netloc.

        See :obj:`~.BaseUrl.parsed_url`.

        Returns:
            str: The url netloc (e.g.: www.example.com:9090)
        """
        return self.parsed_url.netloc

    @property
    def hostname(self):
        """Shortcut for parsed_url.hostname.

        See :obj:`~.BaseUrl.parsed_url`.

        Returns:
            str: The url hostname (e.g.: www.example.com)
        """
        return self.parsed_url.hostname

    @property
    def port(self):
        """Shortcut for parsed_url.port.

        See :obj:`~.BaseUrl.parsed_url`.

        Returns:
            str: The url port (e.g.: 8080, 80, ...)
        """
        return self.parsed_url.port

    def build_absolute_url(self, path_or_url):
        """Build absolute URL from the provided ``path_or_url``.

        If the provided ``path_or_url`` is an URL, it is returned unchanged.
        If the provided ``path_or_url`` is a URL path, it is appended to the base url.

        Examples.::

            BaseUrl('https://example.com').build_absolute_url('/my/path?a=2')
            # -> https://example.com/my/path?a=2

            BaseUrl('https://example.com').build_absolute_url('https://other.example.com/a/b')
            # -> https://other.example.com/a/b

        Args:
            path_or_url (str): URL path or an URL.

        Returns:
            [type]: [description]
        """
        return urllib.parse.urljoin(self.parsed_url.geturl(), path_or_url)

    def __str__(self):
        return self.parsed_url.geturl()



import os
import shutil

from future.utils import python_2_unicode_compatible

from ievv_opensource.utils.ievvbuildstatic.options_mixin import OptionsMixin
from ievv_opensource.utils.ievvbuildstatic.watcher import WatchdogWatchConfig
from ievv_opensource.utils.logmixin import LogMixin


@python_2_unicode_compatible
class Plugin(LogMixin, OptionsMixin):
    """
    Base class for all plugins in ``ievvbuildstatic``.
    """

    #: The name of the plugin.
    name = None

    #: The default group of the plugin
    #: The default value for the ``group`` kwarg for :meth:`.__init__``.
    default_group = None

    def __init__(self, group=None):
        """

        Args:
            group: The group of the plugin. Defaults to :obj:`~.Plugin.default_group`.
                If this is ``None``, the plugin can not be skipped when building
                an app.

                You can use any value for groups, but these groups are
                convenient because the ``ievv buildstatic`` command has
                command line options that makes it easier to skip them:

                - ``"css"``: Should be used for plugins that build css.
                  Skipped when running ``ievv buildstatic``
                  with ``--skip-css``.
                - ``"js"``: Should be used for plugins that build javascript.
                  Skipped when running ``ievv buildstatic``
                  with ``--skip-js``.
                - ``"jstest"``: Should be used for plugins that run automatic tests for javascript code.
                  Should also be used for tests in languages that compile to javascript
                  such as TypeScript and CoffeeScript.
                  Skipped when running ``ievv buildstatic`` with ``--skip-jstests`` or ``--skip-js``.
                - ``"slow-jstest"``: Should be used instead of ``"jstest"`` for very
                  slow tests.
                  Skipped when running ``ievv buildstatic`` with ``--skip-slow-jstests``,
                  ``--skip-jstests`` or ``--skip-js``.
        """
        self.app = None
        self.group = group or self.default_group
        self._temporary_files_and_directories = set()

    def get_loglevel(self):
        return self.app.apps.loglevel

    def get_command_error_message(self):
        return self.app.apps.command_error_message

    def install(self):
        """
        Install any packages required for this plugin.

        Should use :meth:`ievv_opensource.utils.ievvbuild.config.App.get_installer`.

        Examples:

            Install an npm package::

                def install(self):
                    self.app.get_installer('npm').install(
                        'somepackage')
                    self.app.get_installer('npm').install(
                        'otherpackage', version='~1.0.0')
        """

    def post_install(self):
        """
        Called just like install, but after all apps has finished installing.
        """

    def run(self):
        """
        Run the plugin. Put the code executed by the plugin each time files
        change here.
        """
        pass

    def runwrapper(self):
        try:
            self.run()
        except:
            self.after_run_cleanup()
            raise
        else:
            self.after_run_cleanup()

    def after_run_cleanup(self):
        if not self.app.keep_temporary_files:
            self.clean_temporary_files()

    def watch(self):
        """
        Configure watching for this plugin.

        You normally do not override this method, instead you override
        :meth:`.get_watch_folders` and :meth:`.get_watch_regexes`.

        Returns:
            WatchdogWatchConfig: A :class:`ievv_opensource.utils.ievvbuildstatic.watcher.WatchConfig`
                object if you want to watch for changes in this plugin, or ``None`` if you do not
                want to watch for changes.
        """
        watchfolders = self.get_watch_folders()
        if watchfolders:
            watchregexes = self.get_watch_regexes()
            return WatchdogWatchConfig(
                watchfolders=watchfolders,
                watchregexes=watchregexes,
                plugin=self)
        else:
            return None

    def get_watch_regexes(self):
        """
        Get the regex used when watching for changes to files.

        Defaults to a regex matching any files.
        """
        return [r'^.*$']

    def get_watch_folders(self):
        """
        Get folders to watch for changes when using ``ievv buildstatic --watch``.

        Defaults to an empty list, which means that no watching thread is started
        for the plugin.

        The folder paths must be absolute, so in most cases you should
        use ``self.app.get_source_path()`` (see
        :meth:`ievv_opensource.utils.ievvbuildstatic.config.App#get_source_path`)
        to turn user provided relative folder names into absolute paths.
        """
        return []

    def get_logger_name(self):
        return '{}.{}'.format(self.app.get_logger_name(), self.name)

    def __str__(self):
        return self.get_logger_name()

    def make_temporary_build_directory(self):
        """
        Make a temporary directory that you can use for building something.

        Returns:
            str: The absolute path of the new directory.
        """
        path = self.app.make_temporary_build_directory(self.name)
        self.register_temporary_file_or_directory(path)
        # Register the root temporary directory for the app as a temporary directory
        # - we do not get problems with duplicates - the registry is a set.
        self.register_temporary_file_or_directory(self.app.get_temporary_build_directory_path())
        return path

    def register_temporary_file_or_directory(self, absolute_path):
        """
        Register a temporary file or directory.

        This will automatically be cleaned after run() is finished (even if it crashes)
        if the app uses ``keep_temporary_files=False`` (the default).
        """
        self._temporary_files_and_directories.add(absolute_path)

    def clean_temporary_files(self):
        for absolute_path in self._temporary_files_and_directories:
            if os.path.exists(absolute_path):
                if os.path.isdir(absolute_path):
                    shutil.rmtree(absolute_path)
                else:
                    os.remove(absolute_path)

    def add_deferred_success(self, message):
        self.app.add_deferred_success('[{}] {}'.format(self.name, message))

    def add_deferred_warning(self, message):
        self.app.add_deferred_warning('[{}] {}'.format(self.name, message))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2019-03-01 09:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ievv_sms', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='debugsmsmessage',
            name='send_as',
            field=models.CharField(blank=True, default='', max_length=50),
        ),
    ]



import json
import os
from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), 'ievv_opensource', 'version.json')) as f:
    version = json.loads(f.read())


setup(
    name='ievv_opensource',
    description='The opensource modules from the commercial IEVV Django framework.',
    version=version,
    author='Espen Angell Kristiansen, Tor Johansen, Magne Westlie',
    author_email='post@appresso.no',
    license='BSD',
    packages=find_packages(exclude=['manage']),
    install_requires=[
        'sh',
        'Django<2',
        'django-rq<2.2.0',  # django-rq 2.2 removes django 1.x support
        'django_dbdev>=1.0.0',
        'django_cradmin>=5.0.0,<7',
        'watchdog',
        'termcolor',
        'psutil',
        'future',
        'arrow',
    ],
    entry_points={
        'console_scripts': [
            'ievv = ievv_opensource.ievvtasks_common.cli:cli',
        ],
    },
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ]
)



from django.apps import AppConfig
from ievv_opensource.ievv_model_mommy_extras import postgres_field_generators


class ModelMommyExtrasAppConfig(AppConfig):
    name = 'ievv_opensource.ievv_model_mommy_extras'
    verbose_name = "IEVV model mommy extras"

    def ready(self):
        postgres_field_generators.add_to_mommy()



from django import test
from django.core.mail import EmailMultiAlternatives

from ievv_opensource.ievv_developemail.email_backend import DevelopEmailBackend
from ievv_opensource.ievv_developemail.models import DevelopEmail


class TestDevelopEmailBackend(test.TestCase):
    def _create_email_object(self, subject, message, from_email, recipient_list,
                             html_message=None):
        mail = EmailMultiAlternatives(subject, message, from_email, recipient_list)
        if html_message:
            mail.attach_alternative(html_message, 'text/html')
        return mail

    def test_send_single_simple_message(self):
        mail = self._create_email_object(subject='Testsubject', message='Testmessage',
                                         from_email='test@example.com',
                                         recipient_list=['recipient@example.com'])
        DevelopEmailBackend().send_messages([mail])
        self.assertEqual(DevelopEmail.objects.count(), 1)
        developemail = DevelopEmail.objects.first()
        self.assertEqual(developemail.subject, 'Testsubject')
        self.assertEqual(developemail.from_email, 'test@example.com')
        self.assertEqual(developemail.to_emails, 'recipient@example.com')
        self.assertEqual(developemail.plaintext_message, 'Testmessage')

    def test_send_messages_html(self):
        mail = self._create_email_object(subject='Testsubject', message='Testmessage',
                                         from_email='test@example.com',
                                         recipient_list=['recipient@example.com'],
                                         html_message='This is a <strong> message')
        DevelopEmailBackend().send_messages([mail])
        self.assertEqual(DevelopEmail.objects.count(), 1)
        developemail = DevelopEmail.objects.first()
        self.assertEqual(developemail.plaintext_message, 'Testmessage')
        self.assertEqual(developemail.html_message, 'This is a <strong> message')

    def test_send_messages_attachments(self):
        mail = self._create_email_object(subject='Testsubject', message='Testmessage',
                                         from_email='test@example.com',
                                         recipient_list=['recipient@example.com'],
                                         html_message='This is a <strong> message')
        mail.attach(filename='test1.txt', content='testcontent1')
        mail.attach(filename='test2.txt', content='testcontent2')
        DevelopEmailBackend().send_messages([mail])
        self.assertEqual(DevelopEmail.objects.count(), 1)
        developemail = DevelopEmail.objects.first()
        self.assertEqual(developemail.plaintext_message, 'Testmessage')
        self.assertEqual(developemail.html_message, 'This is a <strong> message')
        attachments = developemail.get_attachment_dict(include_payload=True)
        self.assertEqual(dict(attachments), {
            'test1.txt': {
                'filename': 'test1.txt',
                'content_type': 'text/plain',
                'payload': b'testcontent1'
            },
            'test2.txt': {
                'filename': 'test2.txt',
                'content_type': 'text/plain',
                'payload': b'testcontent2'
            }
        })



from django.core.exceptions import ValidationError
from django.test import TestCase
from model_mommy import mommy


class TestTagModel(TestCase):
    def test_clean_invalid_tagtype(self):
        testtag = mommy.prepare('ievv_tagframework.Tag', tagtype='invalidstuff')
        with self.settings(IEVV_TAGFRAMEWORK_SITE_TYPE_CHOICES=[('', ''), ('mytype', 'My Type')]):
            with self.assertRaisesMessage(ValidationError,
                                          "Must be one of: '', 'mytype'. Current value: 'invalidstuff'."):
                testtag.clean()

    def test_clean_valid_tagtype(self):
        testtag = mommy.prepare('ievv_tagframework.Tag', tagtype='mytype')
        with self.settings(IEVV_TAGFRAMEWORK_SITE_TYPE_CHOICES=[('', ''), ('mytype', 'My Type')]):
            testtag.clean()  # No ValidationError



import json
import os

import re
import shutil

import sh
import sys

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = 'Make a source distribution of the codebase.'

    def add_arguments(self, parser):
        parser.add_argument('--keep-old-static', dest='keep_old_static',
                            required=False, action='store_true', default=False,
                            help='Keep old production release static files.')
        parser.add_argument('--dirty-node-modules', dest='dirty_node_modules',
                            required=False, action='store_true', default=False,
                            help='Do not use the --yarn-clean-node-modules argument for ievv buildstatic. '
                                 'Normally not recommended, but useful if you know you have clean node '
                                 'modules with no symlinked packages or other things that may '
                                 'mess up the release.')
        parser.add_argument('--ignore-uncommitted-changes', dest='check_for_uncommitted_changes',
                            required=False, action='store_false', default=True,
                            help='Do not abort if we have uncommitted changes. Mostly only useful when '
                                 'developing this management command.')

    def __log_shell_command_stdout(self, line):
        """
        Called by :meth:`.run_shell_command` each time the shell
        command outputs anything to stdout.
        """
        self.stdout.write(line.rstrip())

    def __log_shell_command_stderr(self, line):
        """
        Called by :meth:`.run_shell_command` each time the shell
        command outputs anything to stderr.
        """
        self.stderr.write(line.rstrip())

    def __run_shell_command(self, executable, args=None, kwargs=None):
        """
        Run a shell command.

        Parameters:
            executable: The name or path of the executable.
            args: List of arguments for the ``sh.Command`` object.
            kwargs: Dict of keyword arguments for the ``sh.Command`` object.

        Raises:
            ShellCommandError: When the command fails. See :class:`.ShellCommandError`.
        """
        command = sh.Command(executable)
        args = args or []
        kwargs = kwargs or {}
        command(*args,
                _out=self.__log_shell_command_stdout,
                _err=self.__log_shell_command_stderr,
                **kwargs)

    def get_version_json_file_path(self):
        try:
            return getattr(settings, 'IEVV_PROJECT_VERSIONJSON_PATH')
        except AttributeError:
            raise CommandError('The IEVV_PROJECT_VERSIONJSON_PATH setting is missing.')

    def get_version_from_version_json(self):
        with open(self.get_version_json_file_path(), 'r') as f:
            version_json = f.read()
        return json.loads(version_json)

    def write_version_json(self, version):
        with open(self.get_version_json_file_path(), 'w') as f:
            f.write(json.dumps(version))
            f.close()

    def get_built_static_directories(self):
        static_directories = []
        for app in settings.IEVVTASKS_BUILDSTATIC_APPS.iterapps():
            if app.appname not in self.__get_ignored_apps():
                static_directory = os.path.join(
                    app.get_appfolder(),
                    'static',
                    app.appname,
                    self.get_release_major_version())
                if os.path.exists(static_directory):
                    static_directories.append(static_directory)
        return static_directories

    def get_release_major_version(self):
        if self.current_version.endswith('alpha'):
            return self.current_version.replace('alpha', '').split('.')[0]
        else:
            raise CommandError(
                'Detecting versions for production releases only work '
                'if the version stored in {} ends with "alpha".'.format(
                    self.get_version_json_file_path()))

    def get_release_version(self):
        major_version = self.get_release_major_version()
        return '{}.0.0'.format(major_version)

    def get_new_version_after_release(self):
        major_version = int(self.get_release_major_version())
        return '{}.0.0alpha'.format(major_version + 1)

    def set_release_version(self):
        self.write_version_json(version=self.get_release_version())

    @property
    def ievv_buildstatic_args(self):
        args = ['buildstatic', '--production']
        if not self.dirty_node_modules:
            args.append('--yarn-clean-node-modules')
        return args

    @property
    def ievv_buildstatic_command_string(self):
        return 'ievv {}'.format(' '.join(self.ievv_buildstatic_args))

    def get_extra_git_add_paths(self):
        return getattr(settings, 'IEVVTASKS_MAKE_SOURCE_DIST_EXTRA_GIT_ADD_PATHS', [])

    def build_release(self):
        self.__run_shell_command('ievv', args=self.ievv_buildstatic_args)
        git_add_args = ['add']
        git_add_args.extend(self.get_built_static_directories())
        git_add_args.append(self.get_version_json_file_path())
        git_add_args.append(self.get_extra_git_add_paths())
        self.__run_shell_command('git', args=git_add_args)

    def commit_release(self):
        self.__run_shell_command('git', args=[
            'commit', '-m', 'Release version: {}'.format(self.get_release_version())
        ])

    def tag_release(self):
        self.__run_shell_command('git', args=[
            'tag', 'v{}'.format(self.get_release_version())
        ])

    def build_and_commit(self):
        self.set_release_version()
        self.build_release()
        self.commit_release()
        self.tag_release()

    def revert_version_change(self):
        version = self.get_new_version_after_release()
        self.write_version_json(version=version)
        self.__run_shell_command('git', args=['add', self.get_version_json_file_path()])
        self.__run_shell_command('git', args=[
            'commit', '-m', 'Set version to: {}'.format(version)
        ])

    def print_preview(self):
        action_list = []
        if not self.keep_old_static:
            paths_to_delete = self.__get_version_build_folders_to_delete()
            action_list.extend([
                'Will delete the following old release folders: \n   - {}'.format(
                    '\n   - '.join(paths_to_delete)
                )
            ])
        action_list.extend([
            'Change {} to {!r}'.format(self.get_version_json_file_path(),
                                       self.get_release_version()),
            self.ievv_buildstatic_command_string,
            'Git commit the files built by buildstatic and tag the commit with v{}'.format(
                self.get_release_version()),
            'python setup.py sdist'
            'Change {} to: {!r}'.format(self.get_version_json_file_path(),
                                        self.get_new_version_after_release()),
            'Git commit the version change.',
        ])
        for action in action_list:
            self.stdout.write('- {}'.format(action))

    def __get_cli_input(self):
        """
        Support python 2 raw input.
        """
        if sys.version_info.major == 2:  # pragma: no cover
            return raw_input
        else:  # pragma: no cover
            return input

    def confirm(self):
        self.stdout.write('Summary of actions to be performed:')
        self.print_preview()
        self.stdout.write('\n')
        self.stdout.write('Are you sure you want to to build a source dist?')
        confirm_text = 'yes'
        cli_input = self.__get_cli_input()
        try:
            return cli_input('[type "{}" to confirm or hit enter to abort] '.format(confirm_text)) == confirm_text
        except KeyboardInterrupt:
            self.stdout.write('\n')
            return False

    def has_git_changes(self):
        return sh.Command('git')('diff-index', 'HEAD').strip() != ''

    def __get_ignored_apps(self):
        return getattr(settings, 'IEVVTASKS_MAKE_SOURCE_DIST_IGNORE_APPS', set())

    def __should_delete_old_build_version_folder(self, dir_name):
        """
        Returns `True` if the regex matches the directory name.

        If the setting ``IEVVTASKS_MAKE_SOURCE_DIST_PRODUCTION_BUILD_VERSION_FOLDER_REGEX`` does
        not exist, we assume all old build folder should be deleted.
        """
        build_version_folder_regex = getattr(
            settings, 'IEVVTASKS_MAKE_SOURCE_DIST_PRODUCTION_BUILD_VERSION_FOLDER_REGEX', r'^\d+(a|alpha)?$')
        if build_version_folder_regex:
            return re.match(build_version_folder_regex, dir_name)
        return False

    def __get_version_build_folders_to_delete(self):
        buildstatic_ignore_apps = self.__get_ignored_apps()
        delete_version_paths = []
        for app in settings.IEVVTASKS_BUILDSTATIC_APPS.iterapps():
            if app.appname in buildstatic_ignore_apps:
                continue
            static_directory = os.path.join(
                app.get_appfolder(),
                'static',
                app.appname)
            if not os.path.exists(static_directory):
                continue
            for dir_name in os.listdir(static_directory):
                if self.__should_delete_old_build_version_folder(dir_name=dir_name):
                    absolute_path = os.path.join(static_directory, dir_name)
                    if os.path.isdir(absolute_path):
                        delete_version_paths.append(absolute_path)
        return delete_version_paths

    def remove_old_static_builds(self):
        """
        Remove old production builds.
        """
        self.stdout.write('\nDeleting old production builds')
        for version_folder_path in self.__get_version_build_folders_to_delete():
            self.stdout.write(' -- Deleting: {}'.format(version_folder_path))
            git = sh.Command('git')
            try:
                git(['rm', '-r', version_folder_path])
            except sh.ErrorReturnCode:
                shutil.rmtree(version_folder_path)

    def make_source_dist(self):
        self.__run_shell_command(sys.executable, args=['setup.py', 'sdist'])

    def handle(self, *args, **options):
        check_for_uncommitted_changes = options['check_for_uncommitted_changes']
        self.keep_old_static = options['keep_old_static']
        self.dirty_node_modules = options['dirty_node_modules']
        self.current_version = None
        self.should_make_source_dist = True

        self.current_version = self.get_version_from_version_json()
        if check_for_uncommitted_changes and self.has_git_changes():
            raise CommandError('You have uncommitted changes. Aborting.')

        if not self.confirm():
            self.stderr.write('Aborted')
            return

        if not self.keep_old_static:
            self.remove_old_static_builds()

        self.current_version = self.get_version_from_version_json()
        self.build_and_commit()

        try:
            self.make_source_dist()
        finally:
            if not self.current_version:
                raise CommandError('Something is wrong - current_version is empty/blank/none')
            self.revert_version_change()
